from __future__ import annotations

import os
from typing import TYPE_CHECKING

import requests
from dotenv import dotenv_values

from bec_lib import MessageEndpoints, RedisConnector, bec_logger, messages
from bec_server.scihub.repeated_timer import RepeatedTimer

logger = bec_logger.logger

if TYPE_CHECKING:
    from bec_server.scihub import SciHub


class SciLogConnector:
    token_expiration_time = 86400  # one day

    def __init__(self, scihub: SciHub, connector: RedisConnector) -> None:
        self.scihub = scihub
        self.connector = connector
        self.host = None
        self.user = None
        self.user_secret = None
        self._configured = False
        self._scilog_thread = None
        self._load_environment()
        self._start_scilog_update()

    def get_token(self) -> str:
        """get a new scilog token"""
        response = requests.post(
            f"{self.host}/users/login",
            json={"principal": self.user, "password": self.user_secret},
            timeout=5,
        )
        if response.ok:
            return response.json()["token"]
        return None

    def set_bec_token(self, token: str) -> None:
        """set the scilog token in redis"""
        self.connector.set(
            MessageEndpoints.logbook(),
            messages.CredentialsMessage(
                credentials={"url": self.host, "token": f"Bearer {token}", "user": self.user}
            ),
        )

    def _start_scilog_update(self) -> None:
        if not self._configured:
            logger.warning("No environment file found. Cannot connect to SciLog.")
            return
        self._scilog_update()
        self._scilog_thread = RepeatedTimer(self.token_expiration_time, self._scilog_update)

    def _scilog_update(self):
        logger.info("Updating SciLog token.")
        token = self.get_token()
        if token:
            self.set_bec_token(token)

    def _load_environment(self):
        env_base = self.scihub.config.service_config.get("scilog", {}).get("env_file")
        if not env_base:
            return
        env_file = os.path.join(env_base, ".env")
        config = dotenv_values(env_file)
        self._update_config(**config)

    def _update_config(
        self,
        SCILOG_DEFAULT_HOST: str = None,
        SCILOG_USER: str = None,
        SCILOG_USER_SECRET: str = None,
        **kwargs,
    ) -> None:
        self.host = SCILOG_DEFAULT_HOST
        self.user = SCILOG_USER
        self.user_secret = SCILOG_USER_SECRET

        if self.host and self.user and self.user_secret:
            self._configured = True

    def shutdown(self):
        if self._scilog_thread:
            self._scilog_thread.stop()



import contextlib
import os
from pathlib import Path
from unittest import mock

import pytest

import bec_lib
from bec_lib import bec_logger, messages
from bec_lib.bec_service import BECService
from bec_lib.endpoints import MessageEndpoints
from bec_lib.messages import BECStatus
from bec_lib.service_config import ServiceConfig

# pylint: disable=no-member
# pylint: disable=missing-function-docstring
# pylint: disable=redefined-outer-name
# pylint: disable=protected-access

dir_path = os.path.dirname(bec_lib.__file__)


@contextlib.contextmanager
def bec_service(config, connector_cls=None, **kwargs):
    if connector_cls is None:
        connector_cls = mock.MagicMock()
    service = BECService(config=config, connector_cls=connector_cls, **kwargs)
    try:
        yield service
    finally:
        service.shutdown()
        bec_logger.logger.remove()
        bec_logger._reset_singleton()


def test_bec_service_init_with_service_config():
    config = ServiceConfig(redis={"host": "localhost", "port": 6379})
    with bec_service(config) as service:
        assert service._service_config == config
        assert service.bootstrap_server == "localhost:6379"
        assert service._unique_service is False


def test_bec_service_init_raises_for_invalid_config():
    with pytest.raises(TypeError):
        with bec_service(mock.MagicMock()):
            ...


def test_bec_service_init_with_service_config_path():
    with bec_service(config=f"{dir_path}/tests/test_service_config.yaml") as service:
        assert isinstance(service._service_config, ServiceConfig)
        assert service.bootstrap_server == "localhost:6379"
        assert service._unique_service is False


def test_init_runs_service_check():
    with mock.patch.object(
        BECService, "_update_existing_services", return_value=False
    ) as mock_update_existing_services:
        with bec_service(f"{dir_path}/tests/test_service_config.yaml", unique_service=True):
            mock_update_existing_services.assert_called_once()


def test_run_service_check_raises_for_existing_service():
    with mock.patch.object(
        BECService, "_update_existing_services", return_value=False
    ) as mock_update_existing_services:
        with bec_service(
            f"{dir_path}/tests/test_service_config.yaml", unique_service=True
        ) as service:
            service._services_info = {"BECService": mock.MagicMock()}
            with pytest.raises(RuntimeError):
                service._run_service_check(timeout_time=0, elapsed_time=10)


def test_run_service_check_repeats():
    with mock.patch.object(
        BECService, "_update_existing_services", return_value=False
    ) as mock_update_existing_services:
        with bec_service(
            f"{dir_path}/tests/test_service_config.yaml", unique_service=True
        ) as service:
            service._services_info = {"BECService": mock.MagicMock()}
            assert service._run_service_check(timeout_time=0.5, elapsed_time=0) is True


def test_bec_service_service_status():
    with mock.patch.object(
        BECService, "_update_existing_services", return_value=False
    ) as mock_update_existing_services:
        with bec_service(
            f"{dir_path}/tests/test_service_config.yaml", unique_service=True
        ) as service:
            mock_update_existing_services.reset_mock()
            status = service.service_status
            mock_update_existing_services.assert_called_once()


def test_bec_service_update_existing_services():
    service_keys = [
        MessageEndpoints.service_status("service1").endpoint.encode(),
        MessageEndpoints.service_status("service2").endpoint.encode(),
    ]
    service_msgs = [
        messages.StatusMessage(name="service1", status=BECStatus.RUNNING, info={}, metadata={}),
        messages.StatusMessage(name="service2", status=BECStatus.IDLE, info={}, metadata={}),
    ]
    connector_cls = mock.MagicMock()
    connector_cls().keys.return_value = service_keys
    connector_cls().get.side_effect = [msg for msg in service_msgs]
    with bec_service(
        f"{os.path.dirname(bec_lib.__file__)}/tests/test_service_config.yaml",
        connector_cls=connector_cls,
        unique_service=True,
    ) as service:
        assert service._services_info == {"service1": service_msgs[0], "service2": service_msgs[1]}


def test_bec_service_update_existing_services_ignores_wrong_msgs():
    service_keys = [
        MessageEndpoints.service_status("service1").endpoint.encode(),
        MessageEndpoints.service_status("service2").endpoint.encode(),
    ]
    service_msgs = [
        messages.StatusMessage(name="service1", status=BECStatus.RUNNING, info={}, metadata={}),
        None,
    ]
    connector_cls = mock.MagicMock()
    connector_cls().keys.return_value = service_keys
    connector_cls().get.side_effect = [service_msgs[0], None]
    with bec_service(
        f"{os.path.dirname(bec_lib.__file__)}/tests/test_service_config.yaml",
        connector_cls=connector_cls,
        unique_service=True,
    ) as service:
        assert service._services_info == {"service1": service_msgs[0]}


def test_bec_service_default_config():
    with bec_service(
        f"{os.path.dirname(bec_lib.__file__)}/tests/test_service_config.yaml", unique_service=True
    ) as service:
        assert service._service_config.service_config["file_writer"]["base_path"] == "./"

    config = ServiceConfig(redis={"host": "localhost", "port": 6379})
    with bec_service(config=config, unique_service=True) as service:
        bec_lib_path = str(Path(bec_lib.service_config.__file__).resolve().parent.parent.parent)
        if "nox" not in bec_lib_path:
            assert (
                os.path.abspath(service._service_config.service_config["file_writer"]["base_path"])
                == bec_lib_path
            )



import builtins
import time

bec = builtins.__dict__.get("bec")
dev = builtins.__dict__.get("dev")
umv = builtins.__dict__.get("umv")


class XTreme:
    def set_au_mesh(self, val: float):
        """Set the position of gold mesh, used for IO

        Args:
            val (float): target position for gold mesh

        Example:
            >>> xtreme.set_au_mesh(163.0)
        """
        umv(dev.gold_mesh1, val)

    def set_slit(self, val: float):
        """Set the exit slit

        Args:
            val (float): width of the exit slit

        Example:
            >>> xtreme.set_slit(30)
        """
        umv(dev.exit_slit, val)

    def set_har(self, val: float):
        """Set the ID harmonics

        Args:
            val (float): Target value

        Example:
            >>> xtreme.set_har(1.0)
        """
        raise NotImplementedError()

    def set_energy(self, val: float):
        """Set the x-ray energy. Internally, it will adjust the id gap and the mono accordingly.

        Args:
            val (float): Target energy in keV

        Example:
            >>> xtreme.set_energy(700.0)
        """
        raise NotImplementedError()

    def open_valve(self, delay=0.85):
        """Open the valve"""
        dev.valve.set(1).wait()
        time.sleep(delay)
        while True:
            valve_val = dev.valve.read()["valve"]["value"]
            if valve_val == 5:
                break
            print(f"Valve did not open. Current status: {valve_val}. Trying again...")
            time.sleep(1)
        print("Valve opened")

    def close_valve(self, delay=0.85):
        """Close the valve"""
        dev.valve.set(0).wait()
        time.sleep(delay)
        while True:
            valve_val = dev.valve.read()["valve"]["value"]
            if valve_val == 2:
                break
            print(f"Valve did not close. Current status: {valve_val}. Trying again...")
            time.sleep(1)
        print("Valve closed")

    def set_hor(self, val: float):
        """Set the horizontal position the endstation.

        Args:
            val (float): Target value

        Example:
            >>> xtreme.set_hor(104.5)
        """
        umv(dev.sample_hor, val)

    def set_vert(self, val: float):
        """Set the vertical position the sample stick.

        Args:
            val (float): Target value

        Example:
            >>> xtreme.set_vert(7.5)
        """
        umv(dev.sample_vert, val)

    def set_hx(self, val: float):
        """Set the magnetic field (hx)

        Args:
            val (float): Target value.

        Example:
            >>> xtreme.set_hx(104.5)
        """
        umv(dev.field_x, val)

    def set_temp(self, val: float):
        """Set the sample temperature.

        Args:
            val (float): Target value.

        Example:
            >>> xtreme.set_temp(300.0)
        """
        umv(dev.temperature, val)

    def set_fe(self, opening: float):
        """Set the FE aperture. Must be either 0, 0.1, 0.25, 0.5, 1, 1.25 or 2.

        Args:
            opening (float): Opening value in mm.

        Example:
            >>> xtreme.set_fe(1)
        """
        # aperture returns an int, not a string?

        # opening *= 1000
        # opening = int(opening)

        # if opening == 0:
        #     dev.

    def wait_temp(self):
        """Check if the 1K pot is refilling and wait if needed."""

        # Not entirely sure what the script is supposed to do...
        raise NotImplementedError()

    def set_range(self):
        # Could not find the keithley pvs...
        raise NotImplementedError()



""" Module to test file_utils.py"""

# pylint: skip-file
import os
from unittest import mock

import pytest

from bec_lib.bec_errors import ServiceConfigError
from bec_lib.endpoints import MessageEndpoints
from bec_lib.file_utils import (
    DeviceConfigWriter,
    FileWriter,
    FileWriterError,
    LogWriter,
    ServiceConfigParser,
)
from bec_lib.messages import ScanStatusMessage
from bec_lib.tests.utils import ConnectorMock


@pytest.fixture(scope="function")
def mock_connector():
    """Mock connector fixture"""
    connector = ConnectorMock("")
    yield connector


@pytest.fixture(scope="function")
def file_writer(mock_connector):
    """File writer fixture"""
    with mock.patch("os.makedirs") as mock_make_dirs:
        yield FileWriter(service_config={"base_path": "/tmp"}, connector=mock_connector)


@pytest.fixture(scope="function")
def scan_msg():
    """Scan message fixture"""
    yield ScanStatusMessage(
        scan_id="1111",
        info={
            "scan_number": 5,
            "file_writer_data": {"file_suffix": "SampleA", "file_directory": "test_dir"},
        },
        status="closed",
    )


def test_device_config_writer():
    """Device config writer fixture and ServiceConfigParser class"""
    with mock.patch("os.makedirs") as mock_make_dirs:
        dcw = DeviceConfigWriter(service_config={"base_path": "/tmp"})
        assert mock_make_dirs.call_count == 1
        assert dcw.directory == "/tmp/device_configs"
        assert dcw.get_recovery_directory() == "/tmp/device_configs/recovery_configs"


def test_log_writer():
    """Device config writer fixture and ServiceConfigParser class"""
    with mock.patch("os.makedirs") as mock_make_dirs:
        lw = LogWriter(service_config={"base_path": "/tmp"})
        assert mock_make_dirs.call_count == 1
        assert lw.directory == "/tmp/logs"


@pytest.mark.parametrize(
    "service_config, connector, raises",
    [
        ({"base_path": "/tmp"}, ConnectorMock(""), False),
        ({"base_path": "/tmp"}, None, False),
        ({"base_path": None}, ConnectorMock(""), True),
    ],
)
def test_file_writer_init(service_config, connector, raises):
    """Test file writer init"""
    if raises:
        with pytest.raises(ServiceConfigError):
            with mock.patch("os.makedirs") as mock_make_dirs:
                fw = FileWriter(service_config=service_config, connector=connector)
    else:
        with mock.patch("os.makedirs") as mock_make_dirs:
            fw = FileWriter(service_config=service_config, connector=ConnectorMock(""))
            if service_config.get("base_path"):
                assert fw._base_path == service_config.get("base_path")
                assert fw._configured is True
            else:
                assert fw._base_path == "."
                assert fw._configured is False


def test_get_scan_msg(file_writer, scan_msg):
    """Test get_scan_msg method"""
    file_writer.connector._get_buffer = {MessageEndpoints.scan_status().endpoint: scan_msg}
    rtr = file_writer.get_scan_msg()
    assert rtr == scan_msg


@pytest.mark.parametrize(
    "scan_number,bundle,lead,user_suffix,ref_path",
    [
        (10, 1000, None, None, "S0000-0999/S0010"),
        (2001, 1000, None, "", "S2000-2999/S2001"),
        (20, 50, 4, "sampleA", "S0000-0049/S0020_sampleA"),
        (20, 50, 5, "sampleB", "S00000-00049/S00020_sampleB"),
        (1200, 1000, 5, "", "S01000-01999/S01200"),
    ],
)
def test_file_writer_get_scan_dir(scan_number, bundle, lead, user_suffix, ref_path, file_writer):

    dir_path = file_writer.get_scan_directory(
        scan_bundle=bundle, scan_number=scan_number, leading_zeros=lead, user_suffix=user_suffix
    )
    assert dir_path == ref_path


def test_compile_full_filename_not_configured(file_writer, scan_msg):
    file_writer.get_scan_msg = mock.MagicMock(return_value=scan_msg)
    file_writer._configured = False
    suffix = "test"
    path = os.path.join(file_writer._base_path, "data", f"S00000_default_{suffix}.h5")
    fp = file_writer.compile_full_filename(suffix=suffix)
    assert fp == path


def test_compile_full_filename(file_writer, scan_msg):
    suffix = "test"
    file_type = ".csv"
    # case 1
    with mock.patch.object(file_writer, "get_scan_msg", return_value=None):
        return_value = file_writer.compile_full_filename(suffix=suffix)
        assert return_value == ""
    # case 2
    with mock.patch.object(file_writer, "get_scan_msg", return_value=scan_msg):
        return_value = file_writer.compile_full_filename(suffix=suffix)
        scannr = scan_msg.info.get("scan_number")
        suffix2 = suffix + f"_{scan_msg.info['file_writer_data']['file_suffix']}"
        expected = os.path.join(
            file_writer._base_path,
            "data",
            scan_msg.info["file_writer_data"]["file_directory"],
            f"S{scannr:0{file_writer._leading_zeros}d}_{suffix2}.h5",
        )
        assert return_value == expected

    # case 3
    scan_msg.info["file_writer_data"].pop("file_suffix")
    scan_msg.info["file_writer_data"].pop("file_directory")
    with mock.patch.object(file_writer, "get_scan_msg", return_value=scan_msg):
        scannr = scan_msg.info.get("scan_number")
        scan_dir = f"S0000-0999/S{scannr:0{file_writer._leading_zeros}d}"
        expected = os.path.join(
            file_writer._base_path,
            "data",
            scan_dir,
            f"S{scannr:0{file_writer._leading_zeros}d}_{suffix}.h5",
        )
        with mock.patch.object(file_writer, "get_scan_directory", return_value=scan_dir):
            return_value = file_writer.compile_full_filename(suffix=suffix)
            assert return_value == expected



"""
This module contains the DeviceManagerDS class, which is a subclass of 
the DeviceManagerBase class and is the main device manager for devices 
in BEC. It is the only place where devices are initialized and managed.
"""

from __future__ import annotations

import inspect
import time
import traceback
from functools import reduce

import numpy as np
import ophyd
import ophyd_devices as opd
from ophyd.ophydobj import OphydObject
from ophyd.signal import EpicsSignalBase
from typeguard import typechecked

from bec_lib import (
    BECService,
    DeviceBase,
    DeviceConfigError,
    DeviceManagerBase,
    MessageEndpoints,
    bec_logger,
    messages,
    plugin_helper,
)
from bec_server.device_server.devices.config_update_handler import ConfigUpdateHandler
from bec_server.device_server.devices.device_serializer import get_device_info

logger = bec_logger.logger


def rgetattr(obj, attr, *args):
    """See https://stackoverflow.com/questions/31174295/getattr-and-setattr-on-nested-objects"""

    def _getattr(obj, attr):
        return getattr(obj, attr, *args)

    return reduce(_getattr, [obj] + attr.split("."))


class DSDevice(DeviceBase):
    def __init__(self, name, obj, config, parent=None):
        super().__init__(name=name, config=config, parent=parent)
        self.obj = obj
        self.metadata = {}
        self.initialized = False

    def __getattr__(self, name: str) -> inspect.Any:
        if hasattr(self.obj, name):
            # compatibility with ophyd devices accessed on the client side
            return rgetattr(self.obj, name)
        return super().__getattr__(name)

    def initialize_device_buffer(self, connector):
        """initialize the device read and readback buffer on redis with a new reading"""
        dev_msg = messages.DeviceMessage(signals=self.obj.read(), metadata={})
        dev_config_msg = messages.DeviceMessage(signals=self.obj.read_configuration(), metadata={})
        if hasattr(self.obj, "low_limit_travel") and hasattr(self.obj, "high_limit_travel"):
            limits = {
                "low": {"value": self.obj.low_limit_travel.get()},
                "high": {"value": self.obj.high_limit_travel.get()},
            }
        else:
            limits = None
        pipe = connector.pipeline()
        connector.set_and_publish(MessageEndpoints.device_readback(self.name), dev_msg, pipe=pipe)
        connector.set_and_publish(
            topic=MessageEndpoints.device_read(self.name), msg=dev_msg, pipe=pipe
        )
        connector.set_and_publish(
            MessageEndpoints.device_read_configuration(self.name), dev_config_msg, pipe=pipe
        )
        if limits is not None:
            connector.set_and_publish(
                MessageEndpoints.device_limits(self.name),
                messages.DeviceMessage(signals=limits),
                pipe=pipe,
            )
        pipe.execute()
        self.initialized = True


class DeviceManagerDS(DeviceManagerBase):
    def __init__(
        self,
        service: BECService,
        config_update_handler: ConfigUpdateHandler = None,
        status_cb: list = None,
    ):
        super().__init__(service, status_cb)
        self._config_request_connector = None
        self._device_instructions_connector = None
        self._config_update_handler_cls = config_update_handler
        self.config_update_handler = None
        self.failed_devices = []

    def initialize(self, bootstrap_server) -> None:
        self.config_update_handler = (
            self._config_update_handler_cls
            if self._config_update_handler_cls is not None
            else ConfigUpdateHandler(device_manager=self)
        )
        super().initialize(bootstrap_server)

    def _reload_action(self) -> None:
        pass

    @staticmethod
    def _get_device_class(dev_type: str) -> type:
        """Get the device class from the device type"""
        return plugin_helper.get_plugin_class(dev_type, [opd, ophyd])

    def _load_session(self, *_args, **_kwargs):
        delayed_init = []
        if not self._is_config_valid():
            self._reset_config()
            return

        try:
            self.failed_devices = {}
            for dev in self._session["devices"]:
                name = dev.get("name")
                enabled = dev.get("enabled")
                logger.info(f"Adding device {name}: {'ENABLED' if enabled else 'DISABLED'}")
                try:
                    dev_cls = self._get_device_class(dev.get("deviceClass"))
                    if issubclass(dev_cls, (opd.DeviceProxy, opd.ComputedSignal)):
                        delayed_init.append(dev)
                        continue
                    self.initialize_device(dev)
                except ConnectionError:
                    msg = traceback.format_exc()
                    self.failed_devices[name] = msg

            for dev in delayed_init:
                try:
                    name = dev.get("name")
                    self.initialize_delayed_devices(dev)
                except (ConnectionError, DeviceConfigError):
                    msg = traceback.format_exc()
                    self.failed_devices[name] = msg
            self.config_update_handler.handle_failed_device_inits()
        except Exception as exc:
            content = traceback.format_exc()
            logger.error(
                f"Failed to initialize device: {dev}: {content}. The config will be reset."
            )
            self._reset_config()
            raise DeviceConfigError(
                f"Failed to initialize device: {dev}: {content}. The config will be reset."
            ) from exc

    def initialize_delayed_devices(self, dev: dict) -> None:
        """Initialize delayed device after all other devices have been initialized."""
        name = dev.get("name")
        enabled = dev.get("enabled")
        logger.info(f"Adding device {name}: {'ENABLED' if enabled else 'DISABLED'}")
        obj = self.initialize_device(dev)

        if hasattr(obj.obj, "lookup"):
            self._register_device_proxy(name)

    def _register_device_proxy(self, name: str) -> None:
        obj_lookup = self.devices.get(name).obj.lookup
        for key in obj_lookup.keys():
            signal_name = obj_lookup[key].get("signal_name")
            if key not in self.devices:
                raise DeviceConfigError(
                    f"Failed to init DeviceProxy {name}, no device {key} found in device manager."
                )
            dev_obj = self.devices[key].obj
            registered_proxies = dev_obj.registered_proxies
            if not hasattr(dev_obj, signal_name):
                raise DeviceConfigError(
                    f"Failed to init DeviceProxy {name}, no signal {signal_name} found for device {key}."
                )
            if key not in registered_proxies:
                # pylint: disable=protected-access
                self.devices[key].obj._registered_proxies.update({name: signal_name})
                continue
            if key in registered_proxies and signal_name not in registered_proxies[key]:
                # pylint: disable=protected-access
                self.devices[key].obj._registered_proxies.update({name: signal_name})
                continue
            if key in registered_proxies.keys() and signal_name in registered_proxies[key]:
                raise RuntimeError(
                    f"Failed to init DeviceProxy {name}, device {key} already has a registered DeviceProxy for {signal_name}. Only one DeviceProxy can be active per signal."
                )

    def _reset_config(self):
        current_config = self._session["devices"]
        if current_config:
            # store the current config in the history
            current_config_msg = messages.AvailableResourceMessage(
                resource=current_config, metadata={"removed_at": time.time()}
            )
            self.producer.lpush(
                MessageEndpoints.device_config_history(), current_config_msg, max_size=50
            )
        msg = messages.AvailableResourceMessage(resource=[])
        self.producer.set(MessageEndpoints.device_config(), msg)
        reload_msg = messages.DeviceConfigMessage(action="reload", config={})
        self.producer.send(MessageEndpoints.device_config_update(), reload_msg)

    @staticmethod
    def update_config(obj: OphydObject, config: dict) -> None:
        """Update an ophyd device's config

        Args:
            obj (Ophydobj): Ophyd object that should be updated
            config (dict): Config dictionary

        """
        if hasattr(obj, "_update_device_config"):
            obj._update_device_config(config)
        else:
            for config_key, config_value in config.items():
                # first handle the ophyd exceptions...
                if config_key == "limits":
                    if hasattr(obj, "low_limit_travel") and hasattr(obj, "high_limit_travel"):
                        low_limit_status = obj.low_limit_travel.set(config_value[0])
                        high_limit_status = obj.high_limit_travel.set(config_value[1])
                        low_limit_status.wait()
                        high_limit_status.wait()
                        continue
                if config_key == "labels":
                    if not config_value:
                        config_value = set()
                    # pylint: disable=protected-access
                    obj._ophyd_labels_ = set(config_value)
                    continue
                if not hasattr(obj, config_key):
                    raise DeviceConfigError(
                        f"Unknown config parameter {config_key} for device of type"
                        f" {obj.__class__.__name__}."
                    )

                config_attr = getattr(obj, config_key)
                if isinstance(config_attr, ophyd.Signal):
                    config_attr.set(config_value)
                elif callable(config_attr):
                    config_attr(config_value)
                else:
                    setattr(obj, config_key, config_value)

    @staticmethod
    def construct_device_obj(dev: dict, device_manager: DeviceManagerDS) -> (OphydObject, dict):
        """
        Construct a device object from a device config dictionary.

        Args:
            dev (dict): device config dictionary
            device_manager (DeviceManagerDS): device manager instance

        Returns:
            (OphydObject, dict): device object and updated config dictionary
        """
        name = dev.get("name")
        dev_cls = DeviceManagerDS._get_device_class(dev["deviceClass"])
        device_config = dev.get("deviceConfig")
        device_config = device_config if device_config is not None else {}
        config = device_config.copy()
        config["name"] = name

        # pylint: disable=protected-access
        device_classes = [dev_cls]
        if issubclass(dev_cls, ophyd.Signal):
            device_classes.append(ophyd.Signal)
        if issubclass(dev_cls, EpicsSignalBase):
            device_classes.append(EpicsSignalBase)
        if issubclass(dev_cls, ophyd.OphydObject):
            device_classes.append(ophyd.OphydObject)

        # get all init parameters of the device class and its parents
        class_params = set()
        for device_class in device_classes:
            class_params.update(inspect.signature(device_class)._parameters)
        class_params_and_config_keys = class_params & config.keys()

        init_kwargs = {key: config.pop(key) for key in class_params_and_config_keys}
        device_access = config.pop("device_access", None)
        if device_access or (device_access is None and config.get("device_mapping")):
            init_kwargs["device_manager"] = device_manager

        signature = inspect.signature(dev_cls)
        if "device_manager" in signature.parameters:
            init_kwargs["device_manager"] = device_manager

        # initialize the device object
        obj = dev_cls(**init_kwargs)
        return obj, config

    def initialize_device(self, dev: dict) -> DSDevice:
        """
        Prepares a device for later usage.
        This includes inspecting the device class signature,
        initializing the object, refreshing the device info and buffer,
        as well as adding subscriptions.
        """
        name = dev.get("name")
        enabled = dev.get("enabled")

        obj, config = self.construct_device_obj(dev, device_manager=self)
        self.update_config(obj, config)

        # refresh the device info
        pipe = self.connector.pipeline()
        self.reset_device_data(obj, pipe)
        self.publish_device_info(obj, pipe)
        pipe.execute()

        # insert the created device obj into the device manager
        opaas_obj = DSDevice(name=name, obj=obj, config=dev, parent=self)

        # pylint:disable=protected-access # this function is shared with clients and it is currently not foreseen that clients add new devices
        self.devices._add_device(name, opaas_obj)

        if not enabled:
            return opaas_obj

        # update device buffer for enabled devices
        # try:
        self.initialize_enabled_device(opaas_obj)
        # pylint:disable=broad-except
        # except Exception:
        #     error_traceback = traceback.format_exc()
        #     logger.error(
        #         f"{error_traceback}. Failed to stage {opaas_obj.name}. The device will be disabled."
        #     )
        #     opaas_obj.enabled = False

        obj = opaas_obj.obj
        # add subscriptions
        if not hasattr(obj, "event_types"):
            return opaas_obj

        if "readback" in obj.event_types:
            obj.subscribe(self._obj_callback_readback, run=opaas_obj.enabled)
        elif "value" in obj.event_types:
            obj.subscribe(self._obj_callback_readback, run=opaas_obj.enabled)

        if "monitor" in obj.event_types:
            obj.subscribe(self._obj_callback_monitor, run=False)
        if "done_moving" in obj.event_types:
            obj.subscribe(self._obj_callback_done_moving, event_type="done_moving", run=False)
        if "flyer" in obj.event_types:
            obj.subscribe(self._obj_flyer_callback, event_type="flyer", run=False)
        if "progress" in obj.event_types:
            obj.subscribe(self._obj_progress_callback, event_type="progress", run=False)
        if hasattr(obj, "motor_is_moving"):
            obj.motor_is_moving.subscribe(self._obj_callback_is_moving, run=opaas_obj.enabled)

        if hasattr(obj, "component_names"):
            for component_name in obj.component_names:
                component = getattr(obj, component_name)
                if not getattr(component, "_auto_monitor", False):
                    continue
                if component.kind in (ophyd.Kind.normal, ophyd.Kind.hinted):
                    component.subscribe(self._obj_callback_readback, run=False)
                elif component.kind == ophyd.Kind.config:
                    component.subscribe(self._obj_callback_configuration, run=False)

        return opaas_obj

    def initialize_enabled_device(self, opaas_obj):
        """connect to an enabled device and initialize the device buffer"""
        self.connect_device(opaas_obj.obj)
        opaas_obj.initialize_device_buffer(self.connector)

    @staticmethod
    def disconnect_device(obj):
        """disconnect from a device"""
        if not obj.connected:
            return
        if hasattr(obj, "controller"):
            obj.controller.off()
            return
        obj.destroy()

    def reset_device(self, obj: DSDevice):
        """reset a device"""
        obj.initialized = False

    @staticmethod
    def connect_device(obj, wait_for_all=False):
        """establish a connection to a device"""
        try:
            if obj.connected:
                return
            if hasattr(obj, "controller"):
                obj.controller.on()
                return
            if hasattr(obj, "wait_for_connection"):
                try:
                    obj.wait_for_connection(all_signals=wait_for_all, timeout=10)
                except TypeError:
                    obj.wait_for_connection(timeout=10)
                return
            logger.error(
                f"Device {obj.name} does not implement the socket controller interface nor"
                " wait_for_connection and cannot be turned on."
            )
            raise ConnectionError(f"Failed to establish a connection to device {obj.name}")
        except Exception:
            error_traceback = traceback.format_exc()
            logger.error(f"{error_traceback}. Failed to connect to {obj.name}.")
            raise ConnectionError(f"Failed to establish a connection to device {obj.name}")

    def publish_device_info(self, obj: OphydObject, pipe=None) -> None:
        """
        Publish the device info to redis. The device info contains
        inter alia the class name, user functions and signals.

        Args:
            obj (_type_): _description_
        """

        interface = get_device_info(obj)
        self.connector.set(
            MessageEndpoints.device_info(obj.name),
            messages.DeviceInfoMessage(device=obj.name, info=interface),
            pipe,
        )

    def reset_device_data(self, obj: OphydObject, pipe=None) -> None:
        """delete all device data and device info"""
        self.connector.delete(MessageEndpoints.device_status(obj.name), pipe)
        self.connector.delete(MessageEndpoints.device_read(obj.name), pipe)
        self.connector.delete(MessageEndpoints.device_info(obj.name), pipe)

    def _obj_callback_readback(self, *_args, obj: OphydObject, **kwargs):
        if not obj.connected:
            return
        name = obj.root.name
        signals = obj.root.read()
        metadata = self.devices.get(obj.root.name).metadata
        dev_msg = messages.DeviceMessage(signals=signals, metadata=metadata)
        pipe = self.connector.pipeline()
        self.connector.set_and_publish(MessageEndpoints.device_readback(name), dev_msg, pipe)
        pipe.execute()

    def _obj_callback_configuration(self, *_args, obj: OphydObject, **kwargs):
        if not obj.connected:
            return
        name = obj.root.name
        signals = obj.root.read_configuration()
        metadata = self.devices.get(obj.root.name).metadata
        dev_msg = messages.DeviceMessage(signals=signals, metadata=metadata)
        pipe = self.connector.pipeline()
        self.connector.set_and_publish(
            MessageEndpoints.device_read_configuration(name), dev_msg, pipe
        )
        pipe.execute()

    @typechecked
    def _obj_callback_monitor(self, *_args, obj: OphydObject, value: np.ndarray, **kwargs):
        """
        Callback for ophyd monitor events. Sends the data to redis.
        Introduces a check of the data size, and incoporates a limit which is defined in max_size (in MB)

        Args:
            obj (OphydObject): ophyd object
            value (np.ndarray): data from ophyd device

        """
        # Convert sizes from bytes to MB
        dsize = len(value.tobytes()) / 1e6
        max_size = 1000
        if dsize > max_size:
            logger.warning(
                f"Data size of single message is too large to send, current max_size {max_size}."
            )
            return
        if obj.connected:
            name = obj.root.name
            metadata = self.devices[name].metadata
            msg = messages.DeviceMonitorMessage(device=name, data=value, metadata=metadata)
            stream_msg = {"data": msg}
            self.connector.xadd(
                MessageEndpoints.device_monitor(name),
                stream_msg,
                max_size=min(100, int(max_size // dsize)),
            )

    def _obj_callback_acq_done(self, *_args, **kwargs):
        device = kwargs["obj"].root.name
        status = 0
        metadata = self.devices[device].metadata
        self.connector.set(
            MessageEndpoints.device_status(device),
            messages.DeviceStatusMessage(device=device, status=status, metadata=metadata),
        )

    def _obj_callback_done_moving(self, *args, **kwargs):
        self._obj_callback_readback(*args, **kwargs)
        # self._obj_callback_acq_done(*args, **kwargs)

    def _obj_callback_is_moving(self, *_args, **kwargs):
        device = kwargs["obj"].root.name
        status = int(kwargs.get("value"))
        metadata = self.devices[device].metadata
        self.connector.set(
            MessageEndpoints.device_status(device),
            messages.DeviceStatusMessage(device=device, status=status, metadata=metadata),
        )

    def _obj_flyer_callback(self, *_args, **kwargs):
        obj = kwargs["obj"]
        data = kwargs["value"].get("data")
        ds_obj = self.devices[obj.root.name]
        metadata = ds_obj.metadata
        if "scan_id" not in metadata:
            return

        if not hasattr(ds_obj, "emitted_points"):
            ds_obj.emitted_points = {}

        emitted_points = ds_obj.emitted_points.get(metadata["scan_id"], 0)

        # make sure all arrays are of equal length
        max_points = min(len(d) for d in data.values())
        bundle = messages.BundleMessage()
        for ii in range(emitted_points, max_points):
            timestamp = time.time()
            signals = {}
            for key, val in data.items():
                signals[key] = {"value": val[ii], "timestamp": timestamp}
            bundle.append(
                messages.DeviceMessage(signals=signals, metadata={"point_id": ii, **metadata})
            )
        ds_obj.emitted_points[metadata["scan_id"]] = max_points
        pipe = self.connector.pipeline()
        self.connector.send(MessageEndpoints.device_read(obj.root.name), bundle, pipe=pipe)
        msg = messages.DeviceStatusMessage(
            device=obj.root.name, status=max_points, metadata=metadata
        )
        self.connector.set_and_publish(
            MessageEndpoints.device_progress(obj.root.name), msg, pipe=pipe
        )
        pipe.execute()

    def _obj_progress_callback(self, *_args, obj, value, max_value, done, **kwargs):
        metadata = self.devices[obj.root.name].metadata
        msg = messages.ProgressMessage(
            value=value, max_value=max_value, done=done, metadata=metadata
        )
        self.connector.set_and_publish(MessageEndpoints.device_progress(obj.root.name), msg)



from unittest import mock

from bec_server.bec_server_utils.service_handler import ServiceHandler


def test_service_handler():
    bec_path = "/path/to/bec"
    config_path = "/path/to/config"

    with mock.patch("bec_server.bec_server_utils.service_handler.sys") as mock_sys:
        mock_sys.platform = "linux"
        service_handler = ServiceHandler(bec_path, config_path)
        assert service_handler.interface == "tmux"


def test_service_handler_start():
    bec_path = "/path/to/bec"
    config_path = "/path/to/config"

    with mock.patch("bec_server.bec_server_utils.service_handler.sys") as mock_sys:
        mock_sys.platform = "linux"
        service_handler = ServiceHandler(bec_path, config_path)

        with mock.patch(
            "bec_server.bec_server_utils.service_handler.tmux_start"
        ) as mock_tmux_start:
            service_handler.start()
            mock_tmux_start.assert_called_once_with(bec_path, config_path, service_handler.SERVICES)


def test_service_handler_stop():
    with mock.patch("bec_server.bec_server_utils.service_handler.tmux_stop") as mock_tmux_stop:
        service_handler = ServiceHandler("/path/to/bec", "/path/to/config")
        service_handler.stop()
        mock_tmux_stop.assert_called_once()


def test_service_handler_restart():
    bec_path = "/path/to/bec"
    config_path = "/path/to/config"

    with mock.patch("bec_server.bec_server_utils.service_handler.sys") as mock_sys:
        mock_sys.platform = "linux"
        service_handler = ServiceHandler(bec_path, config_path)

        with mock.patch("bec_server.bec_server_utils.service_handler.tmux_stop") as mock_tmux_stop:
            with mock.patch(
                "bec_server.bec_server_utils.service_handler.tmux_start"
            ) as mock_tmux_start:
                service_handler.restart()
                mock_tmux_stop.assert_called_once()
                mock_tmux_start.assert_called_once_with(
                    bec_path, config_path, service_handler.SERVICES
                )


def test_service_handler_services():
    service_handler = ServiceHandler("/path/to/bec", "/path/to/config")
    assert (
        service_handler.SERVICES["scan_server"]["path"].substitute(base_path="/path/to/bec")
        == "/path/to/bec/scan_server"
    )

    assert service_handler.SERVICES["scan_server"]["command"] == "bec-scan-server"



"""
BECClient class. This class is the main entry point for the BEC client and all 
derived classes. It is used to initialize the client and start the client.
"""

from __future__ import annotations, print_function

import builtins
import getpass
import importlib
import inspect
from typing import TYPE_CHECKING

from rich.console import Console
from rich.table import Table

from bec_lib import messages
from bec_lib.alarm_handler import AlarmHandler, Alarms
from bec_lib.bec_service import BECService
from bec_lib.bl_checks import BeamlineChecks
from bec_lib.callback_handler import CallbackHandler, EventType
from bec_lib.dap_plugins import DAPPlugins
from bec_lib.devicemanager import DeviceManagerBase
from bec_lib.endpoints import MessageEndpoints
from bec_lib.logger import bec_logger
from bec_lib.redis_connector import RedisConnector
from bec_lib.scan_manager import ScanManager
from bec_lib.scans import Scans
from bec_lib.service_config import ServiceConfig
from bec_lib.user_scripts_mixin import UserScriptsMixin

if TYPE_CHECKING:
    from bec_lib.connector import ConnectorBase

logger = bec_logger.logger


class BECClient(BECService, UserScriptsMixin):
    """
    The BECClient class is the main entry point for the BEC client and all derived classes.
    """

    _client = None
    _initialized = False
    started = False

    def __init__(
        self,
        config: ServiceConfig = None,
        connector_cls: ConnectorBase = None,
        wait_for_server=False,
        forced=False,
        parent=None,
    ) -> None:
        """
        Initialize the BECClient

        Args:
            config (ServiceConfig, optional): The configuration for the client. Defaults to None.
            connector_cls (ConnectorBase, optional): The connector class to use. Defaults to None.
            wait_for_server (bool, optional): Whether to wait for the server to be available before starting. Defaults to False.
            forced (bool, optional): Whether to force the initialization of a new client. Defaults to False.
        """
        if self._initialized:
            return
        self.__init_params = {
            "config": config if config is not None else ServiceConfig(),
            "connector_cls": connector_cls if connector_cls is not None else RedisConnector,
            "wait_for_server": wait_for_server,
        }
        self.device_manager = None
        self.queue = None
        self.alarm_handler = None
        self.config = None
        self.history = None
        self.live_updates = None
        self.dap = None
        self.bl_checks = None
        self._hli_funcs = {}
        self.metadata = {}
        self.file_writer_data = {}
        self.callbacks = CallbackHandler()
        self._parent = parent if parent is not None else self
        self._initialized = True

    def __new__(cls, *args, forced=False, **kwargs):
        if forced or cls._client is None:
            cls._client = super(BECClient, cls).__new__(cls)
            cls._initialized = False
        return cls._client

    def __str__(self) -> str:
        return "BECClient\n\nTo get a list of available commands, type `bec.show_all_commands()`"

    @classmethod
    def _reset_singleton(cls):
        cls._client = None
        cls._initialized = False
        cls.started = False

    @property
    def username(self) -> str:
        """get the current username"""
        return self._username

    @property
    def active_account(self) -> str:
        """get the currently active target (e)account"""
        msg = self.connector.get(MessageEndpoints.account())
        if msg:
            return msg.value
        return ""

    def start(self):
        """start the client"""
        if self.started:
            return
        self.started = True
        config = self.__init_params["config"]
        connector_cls = self.__init_params["connector_cls"]
        wait_for_server = self.__init_params["wait_for_server"]
        super().__init__(config, connector_cls, wait_for_server=wait_for_server)
        builtins.bec = self._parent
        self._start_services()
        logger.info("Starting new client")

    def _start_services(self):
        self._configure_logger()
        self._load_scans()
        # self.logbook = LogbookConnector(self.connector)
        self._update_username()
        self._start_device_manager()
        self._start_scan_queue()
        self._start_alarm_handler()
        self.load_all_user_scripts()
        self.config = self.device_manager.config_helper
        self.history = self.queue.queue_storage.storage
        self.dap = DAPPlugins(self)
        self.bl_checks = BeamlineChecks(self)
        self.bl_checks.start()

    def alarms(self, severity=Alarms.WARNING):
        """get the next alarm with at least the specified severity"""
        if self.alarm_handler is None:
            yield []
        yield from self.alarm_handler.get_alarm(severity=severity)

    def show_all_alarms(self, severity=Alarms.WARNING):
        """print all unhandled alarms"""
        alarms = self.alarm_handler.get_unhandled_alarms(severity=severity)
        for alarm in alarms:
            print(alarm)

    def clear_all_alarms(self):
        """remove all alarms from stack"""
        self.alarm_handler.clear()

    @property
    def pre_scan_macros(self):
        """currently stored pre-scan macros"""
        return self.connector.lrange(MessageEndpoints.pre_scan_macros(), 0, -1)

    @pre_scan_macros.setter
    def pre_scan_macros(self, hooks: list[str]):
        self.connector.delete(MessageEndpoints.pre_scan_macros())
        for hook in hooks:
            msg = messages.VariableMessage(value=hook)
            self.connector.lpush(MessageEndpoints.pre_scan_macros(), msg)

    def _load_scans(self):
        self.scans = Scans(self._parent)
        builtins.__dict__["scans"] = self.scans

    def load_high_level_interface(self, module_name: str) -> None:
        """Load a high level interface module.
        Runs a callback of type `EventType.NAMESPACE_UPDATE`
        to inform clients about added objects in the namesapce.

        Args:
            module_name (str): The name of the module to load
        """
        mod = importlib.import_module(f"bec_ipython_client.high_level_interfaces.{module_name}")
        members = inspect.getmembers(mod)
        funcs = {name: func for name, func in members if not name.startswith("__")}
        self._hli_funcs = funcs
        builtins.__dict__.update(funcs)
        self.callbacks.run(EventType.NAMESPACE_UPDATE, action="add", ns_objects=funcs)

    def _update_username(self):
        self._username = getpass.getuser()

    def _start_scan_queue(self):
        self.queue = ScanManager(self.connector)

    def _configure_logger(self):
        bec_logger.logger.remove()
        bec_logger.add_file_log(bec_logger.LOGLEVEL.DEBUG)
        bec_logger.add_sys_stderr(bec_logger.LOGLEVEL.SUCCESS)
        bec_logger.add_console_log()

    def _start_device_manager(self):
        logger.info("Starting device manager")
        self.device_manager = DeviceManagerBase(self)
        self.device_manager.initialize(self.bootstrap_server)
        builtins.dev = self.device_manager.devices

    def _start_alarm_handler(self):
        logger.info("Starting alarm listener")
        self.alarm_handler = AlarmHandler(self.connector)
        self.alarm_handler.start()

    def shutdown(self):
        """shutdown the client and all its components"""
        if self.started:
            super().shutdown()
        if self.device_manager:
            self.device_manager.shutdown()
        if self.queue:
            self.queue.shutdown()
        if self.alarm_handler:
            self.alarm_handler.shutdown()
        if self.bl_checks:
            self.bl_checks.stop()
        bec_logger.logger.remove()
        self.started = False

    def _print_available_commands(self, title: str, data: tuple) -> None:
        console = Console()
        table = Table(title=title)
        table.add_column("Name", justify="center")
        table.add_column("Description", justify="center")
        for name, descr in data:
            table.add_row(name, descr)
        console.print(table)

    def _print_user_script_commands(self) -> None:
        data = self._get_user_script_commands()
        self._print_available_commands("User scripts", data)

    def _get_user_script_commands(self) -> list:
        avail_commands = []
        for name, val in self._scripts.items():
            descr = self._get_description_from_doc_string(val["cls"].__doc__)
            avail_commands.append((name, descr))
        return avail_commands

    def _get_scan_commands(self) -> list:
        avail_commands = []
        for name, scan in self.scans._available_scans.items():
            descr = self._get_description_from_doc_string(scan.scan_info["doc"])
            avail_commands.append((name, descr))
        return avail_commands

    def _print_scan_commands(self) -> None:
        data = self._get_scan_commands()
        self._print_available_commands("Scans", data)

    def show_all_commands(self):
        self._print_user_script_commands()
        self._print_scan_commands()

    @staticmethod
    def _get_description_from_doc_string(doc_string: str) -> str:
        if not doc_string:
            return ""
        return doc_string.strip().split("\n")[0]



"""
This module provides a class to validate the device configuration against the openapi schema of SciBec.
"""

import json
import os

import fastjsonschema

import bec_lib


class SciBecValidator:
    def __init__(self, schema_path: str = None) -> None:
        if not schema_path:
            schema_path = f"{os.path.dirname(bec_lib.__file__)}/configs/openapi_schema.json"
        self.schema = self._load_schema(schema_path)
        self.device_schema = self.schema["components"]["schemas"]["Device"]
        self.device_validate = fastjsonschema.compile(self.device_schema)

    def _load_schema(self, schema_path) -> dict:
        with open(schema_path, "r", encoding="utf-8") as schema_file:
            content = schema_file.read()
            schema_content = json.loads(content)
        return schema_content

    def validate_device(self, device):
        self.device_validate(device)

    def validate_device_patch(self, config):
        properties = {k: v for k, v in self.device_schema["properties"].items() if k in config}
        schema = {"properties": properties, "required": []}
        fastjsonschema.validate(schema, config)



"""
Post startup script for the BEC client. This script is executed after the
IPython shell is started. It is used to load the beamline specific
information and to setup the prompts.

The script is executed in the global namespace of the IPython shell. This
means that all variables defined here are available in the shell.

While command-line arguments have to be set in the pre-startup script, the
post-startup script can be used to load beamline specific information and
to setup the prompts.

    from bec_lib import bec_logger

    logger = bec_logger.logger

    # pylint: disable=import-error
    _args = _main_dict["args"]

    _session_name = "cSAXS"
    if _args.session.lower() == "lamni":
        from csaxs_bec.bec_ipython_client.plugins.cSAXS import *
        from csaxs_bec.bec_ipython_client.plugins.LamNI import *

        _session_name = "LamNI"
        lamni = LamNI(bec)
        logger.success("LamNI session loaded.")

    elif _args.session.lower() == "csaxs":
        print("Loading cSAXS session")
        from csaxs_bec.bec_ipython_client.plugins.cSAXS import *

        logger.success("cSAXS session loaded.")
"""

# pylint: disable=invalid-name, unused-import, import-error, undefined-variable, unused-variable, unused-argument, no-name-in-module



# pylint: skip-file
from collections import namedtuple
from unittest import mock

import pytest
from ophyd import Device, Kind, Signal, Staged, StatusBase

from bec_lib import Alarms, MessageEndpoints, messages
from bec_server.device_server.rpc_mixin import RPCMixin


@pytest.fixture
def rpc_cls():
    rpc_mixin = RPCMixin()
    rpc_mixin.connector = mock.MagicMock()
    rpc_mixin.connector = mock.MagicMock()
    rpc_mixin.device_manager = mock.MagicMock()
    yield rpc_mixin


@pytest.fixture
def instr():
    yield messages.DeviceInstructionMessage(
        device="device",
        action="rpc",
        parameter={"rpc_id": "rpc_id", "func": "trigger"},
        metadata={"RID": "RID"},
    )


@pytest.mark.parametrize(
    "instr_params",
    [
        ({"args": (1, 2, 3), "kwargs": {"a": 1, "b": 2}}),
        ({"args": (1, 2, 3)}),
        ({"kwargs": {"a": 1, "b": 2}}),
        ({}),
    ],
)
def test_get_result_from_rpc(rpc_cls, instr_params):
    rpc_var = mock.MagicMock()
    rpc_var.return_value = 1
    out = rpc_cls._get_result_from_rpc(rpc_var=rpc_var, instr_params=instr_params)
    if instr_params:
        if instr_params.get("args") and instr_params.get("kwargs"):
            rpc_var.assert_called_once_with(*instr_params["args"], **instr_params["kwargs"])
        elif instr_params.get("args"):
            rpc_var.assert_called_once_with(*instr_params["args"])
        elif instr_params.get("kwargs"):
            rpc_var.assert_called_once_with(**instr_params["kwargs"])
        else:
            rpc_var.assert_called_once_with()
    assert out == 1


@pytest.mark.parametrize("instr_params", [({"args": (1, 2, 3), "kwargs": {"a": 1, "b": 2}}), ({})])
def test_get_result_from_rpc_var(rpc_cls, instr_params):
    rpc_var = 5
    out = rpc_cls._get_result_from_rpc(rpc_var=rpc_var, instr_params=instr_params)
    assert out == 5


def test_get_result_from_rpc_not_serializable(rpc_cls):
    rpc_var = mock.MagicMock()
    rpc_var.return_value = mock.MagicMock()
    rpc_var.return_value.__str__.side_effect = Exception
    out = rpc_cls._get_result_from_rpc(rpc_var=rpc_var, instr_params={})
    assert out is None
    rpc_cls.connector.raise_alarm.assert_called_once_with(
        severity=Alarms.WARNING,
        alarm_type="TypeError",
        source={},
        msg="Return value of rpc call {} is not serializable.",
        metadata={},
    )


def test_get_result_from_rpc_ophyd_status(rpc_cls):
    rpc_var = mock.MagicMock()
    status = StatusBase()
    rpc_var.return_value = status
    out = rpc_cls._get_result_from_rpc(rpc_var=rpc_var, instr_params={})
    assert out is rpc_var.return_value
    status.set_finished()


def test_get_result_from_rpc_list_from_stage(rpc_cls):
    rpc_var = mock.MagicMock()
    rpc_var.return_value = [mock.MagicMock(), mock.MagicMock()]
    rpc_var.return_value[0]._staged = True
    rpc_var.return_value[1]._staged = False
    out = rpc_cls._get_result_from_rpc(rpc_var=rpc_var, instr_params={"func": "stage"})
    assert out == [True, False]


def test_send_rpc_exception(rpc_cls, instr):
    rpc_cls._send_rpc_exception(Exception(), instr)
    rpc_cls.connector.set.assert_called_once_with(
        MessageEndpoints.device_rpc("rpc_id"),
        messages.DeviceRPCMessage(
            device="device",
            return_val=None,
            out={"error": "Exception", "msg": (), "traceback": "NoneType: None\n"},
            success=False,
        ),
    )


def test_send_rpc_result_to_client(rpc_cls):
    result = mock.MagicMock()
    result.getvalue.return_value = "result"
    rpc_cls._send_rpc_result_to_client("device", {"rpc_id": "rpc_id"}, 1, result)
    rpc_cls.connector.set.assert_called_once_with(
        MessageEndpoints.device_rpc("rpc_id"),
        messages.DeviceRPCMessage(device="device", return_val=1, out="result", success=True),
        expire=1800,
    )


def test_run_rpc(rpc_cls, instr):
    rpc_cls._assert_device_is_enabled = mock.MagicMock()
    with (
        mock.patch.object(rpc_cls, "_process_rpc_instruction") as _process_rpc_instruction,
        mock.patch.object(rpc_cls, "_send_rpc_result_to_client") as _send_rpc_result_to_client,
    ):
        _process_rpc_instruction.return_value = 1
        rpc_cls.run_rpc(instr)
        rpc_cls._assert_device_is_enabled.assert_called_once_with(instr)
        _process_rpc_instruction.assert_called_once_with(instr)
        _send_rpc_result_to_client.assert_called_once_with(
            "device", {"rpc_id": "rpc_id", "func": "trigger"}, 1, mock.ANY
        )


def test_run_rpc_sends_rpc_exception(rpc_cls, instr):
    rpc_cls._assert_device_is_enabled = mock.MagicMock()
    with (
        mock.patch.object(rpc_cls, "_process_rpc_instruction") as _process_rpc_instruction,
        mock.patch.object(rpc_cls, "_send_rpc_exception") as _send_rpc_exception,
    ):
        _process_rpc_instruction.side_effect = Exception
        rpc_cls.run_rpc(instr)
        rpc_cls._assert_device_is_enabled.assert_called_once_with(instr)
        _process_rpc_instruction.assert_called_once_with(instr)
        _send_rpc_exception.assert_called_once_with(mock.ANY, instr)


@pytest.fixture()
def dev_mock():
    dev_mock = mock.MagicMock()
    dev_mock.obj = mock.MagicMock(spec=Device)
    dev_mock.obj.readback = mock.MagicMock(spec=Signal)
    dev_mock.obj.readback.kind = Kind.hinted
    dev_mock.obj.user_setpoint = mock.MagicMock(spec=Signal)
    dev_mock.obj.user_setpoint.kind = Kind.normal
    dev_mock.obj.velocity = mock.MagicMock(spec=Signal)
    dev_mock.obj.velocity.kind = Kind.config
    dev_mock.obj.notused = mock.MagicMock(spec=Signal)
    dev_mock.obj.notused.kind = Kind.omitted
    return dev_mock


@pytest.mark.parametrize(
    "func, read_called",
    [
        ("read", True),
        ("read_configuration", False),
        ("readback.read_configuration", False),
        ("readback.read", True),
        ("user_setpoint.read", True),
        ("user_setpoint.read_configuration", False),
        ("velocity.read", False),
        ("velocity.read_configuration", False),
        ("notused.read", False),
        ("notused.read_configuration", False),
    ],
)
def test_process_rpc_instruction_read(rpc_cls, dev_mock, instr, func, read_called):
    instr.content["parameter"]["func"] = func
    rpc_cls.device_manager.devices = {"device": dev_mock}
    rpc_cls._read_and_update_devices = mock.MagicMock()
    rpc_cls._read_config_and_update_devices = mock.MagicMock()
    rpc_cls._process_rpc_instruction(instr)
    if read_called:
        rpc_cls._read_and_update_devices.assert_called_once_with(["device"], instr.metadata)
        rpc_cls._read_config_and_update_devices.assert_not_called()
    else:
        rpc_cls._read_and_update_devices.assert_not_called()
        if "notused" not in func:
            rpc_cls._read_config_and_update_devices.assert_called_once_with(
                ["device"], instr.metadata
            )


def test_process_rpc_instruction_with_status_return(rpc_cls, dev_mock, instr):
    rpc_cls.device_manager.devices = {"device": dev_mock}
    rpc_cls._status_callback = mock.MagicMock()
    with mock.patch.object(rpc_cls, "_get_result_from_rpc") as rpc_result:
        status = StatusBase()
        rpc_result.return_value = status
        res = rpc_cls._process_rpc_instruction(instr)
        assert res == {
            "type": "status",
            "RID": "RID",
            "success": status.success,
            "timeout": status.timeout,
            "done": status.done,
            "settle_time": status.settle_time,
        }
        status.set_finished()


def test_process_rpc_instruction_with_namedtuple_return(rpc_cls, dev_mock, instr):
    rpc_cls.device_manager.devices = {"device": dev_mock}
    with mock.patch.object(rpc_cls, "_get_result_from_rpc") as rpc_result:
        point_type = namedtuple("Point", ["x", "y"])
        point_tuple = point_type(5, 6)
        rpc_result.return_value = point_tuple
        res = rpc_cls._process_rpc_instruction(instr)
        assert res == {
            "type": "namedtuple",
            "RID": instr.metadata.get("RID"),
            "fields": point_tuple._fields,
            "values": point_tuple._asdict(),
        }


@pytest.mark.parametrize(
    "return_val,result",
    [([], []), ([1, 2, 3], [1, 2, 3]), ([Staged.no, Staged.yes], ["Staged.no", "Staged.yes"])],
)
def test_process_rpc_instruction_with_list_return(rpc_cls, dev_mock, instr, return_val, result):
    rpc_cls.device_manager.devices = {"device": dev_mock}
    with mock.patch.object(rpc_cls, "_get_result_from_rpc") as rpc_result:
        rpc_result.return_value = return_val
        res = rpc_cls._process_rpc_instruction(instr)
        assert res == result


def test_process_rpc_instruction_set_attribute(rpc_cls, dev_mock, instr):
    instr.content["parameter"]["kwargs"] = {"_set_property": True}
    instr.content["parameter"]["args"] = [5]
    instr.content["parameter"]["func"] = "attr_value"
    rpc_cls.device_manager.devices = {"device": dev_mock}
    rpc_cls._process_rpc_instruction(instr)
    rpc_cls.device_manager.devices["device"].obj.attr_value == 5


def test_process_rpc_instruction_set_attribute_on_sub_device(rpc_cls, dev_mock, instr):
    instr.content["parameter"]["kwargs"] = {"_set_property": True}
    instr.content["parameter"]["args"] = [5]
    instr.content["parameter"]["func"] = "user_setpoint.attr_value"
    rpc_cls.device_manager.devices = {"device": dev_mock}
    rpc_cls._process_rpc_instruction(instr)
    rpc_cls.device_manager.devices["device"].obj.user_setpoint.attr_value == 5



from ophyd import ADComponent as ADCpt
from ophyd import CamBase, DetectorBase, EpicsSignal, EpicsSignalRO, EpicsSignalWithRBV
from ophyd.areadetector.plugins import FileBase


class slsDetectorCam(CamBase, FileBase):
    detector_type = ADCpt(EpicsSignalRO, "DetectorType_RBV")
    setting = ADCpt(EpicsSignalWithRBV, "Setting")
    delay_time = ADCpt(EpicsSignalWithRBV, "DelayTime")
    threshold_energy = ADCpt(EpicsSignalWithRBV, "ThresholdEnergy")
    enable_trimbits = ADCpt(EpicsSignalWithRBV, "Trimbits")
    bit_depth = ADCpt(EpicsSignalWithRBV, "BitDepth")
    num_gates = ADCpt(EpicsSignalWithRBV, "NumGates")
    num_cycles = ADCpt(EpicsSignalWithRBV, "NumCycles")
    num_frames = ADCpt(EpicsSignalWithRBV, "NumFrames")
    timing_mode = ADCpt(EpicsSignalWithRBV, "TimingMode")
    trigger_software = ADCpt(EpicsSignal, "TriggerSoftware")
    high_voltage = ADCpt(EpicsSignalWithRBV, "HighVoltage")
    # Receiver and data callback
    receiver_mode = ADCpt(EpicsSignalWithRBV, "ReceiverMode")
    receiver_stream = ADCpt(EpicsSignalWithRBV, "ReceiverStream")
    enable_data = ADCpt(EpicsSignalWithRBV, "UseDataCallback")
    missed_packets = ADCpt(EpicsSignalRO, "ReceiverMissedPackets_RBV")
    # Direct settings access
    setup_file = ADCpt(EpicsSignal, "SetupFile")
    load_setup = ADCpt(EpicsSignal, "LoadSetup")
    command = ADCpt(EpicsSignal, "Command")
    # Mythen 3
    counter_mask = ADCpt(EpicsSignalWithRBV, "CounterMask")
    counter1_threshold = ADCpt(EpicsSignalWithRBV, "Counter1Threshold")
    counter2_threshold = ADCpt(EpicsSignalWithRBV, "Counter2Threshold")
    counter3_threshold = ADCpt(EpicsSignalWithRBV, "Counter3Threshold")
    gate1_delay = ADCpt(EpicsSignalWithRBV, "Gate1Delay")
    gate1_width = ADCpt(EpicsSignalWithRBV, "Gate1Width")
    gate2_delay = ADCpt(EpicsSignalWithRBV, "Gate2Delay")
    gate2_width = ADCpt(EpicsSignalWithRBV, "Gate2Width")
    gate3_delay = ADCpt(EpicsSignalWithRBV, "Gate3Delay")
    gate3_width = ADCpt(EpicsSignalWithRBV, "Gate3Width")
    # Moench
    json_frame_mode = ADCpt(EpicsSignalWithRBV, "JsonFrameMode")
    json_detector_mode = ADCpt(EpicsSignalWithRBV, "JsonDetectorMode")


class slsDetector(DetectorBase):
    cam = ADCpt(slsDetectorCam, "cam1:")



import enum
import time
import warnings
from copy import deepcopy
from typing import Any, ClassVar, Literal

import numpy as np
from pydantic import BaseModel, Field, field_validator, model_validator


class BECStatus(enum.Enum):
    """BEC status enum"""

    RUNNING = 2
    BUSY = 1
    IDLE = 0
    ERROR = -1


class BECMessage(BaseModel):
    """Base Model class for BEC Messages

    Args:
        msg_type (str): ClassVar for the message type, subclasses should override this.
        metadata (dict, optional): Optional dictionary with metadata for the BECMessage

    """

    msg_type: ClassVar[str]
    metadata: dict | None = Field(default_factory=dict)

    @field_validator("metadata")
    @classmethod
    def check_metadata(cls, v):
        """Validate the metadata, return empty dict if None

        Args:
            v (dict, None): Metadata dictionary
        """
        if v is None:
            return {}
        return v

    @property
    def content(self):
        """Return the content of the message"""
        content = self.__dict__.copy()
        content.pop("metadata", None)
        return content

    def __eq__(self, other):
        if not isinstance(other, BECMessage):
            # don't attempt to compare against unrelated types
            return False

        try:
            np.testing.assert_equal(self.model_dump(), other.model_dump())
        except AssertionError:
            return False

        return self.msg_type == other.msg_type and self.metadata == other.metadata

    def loads(self):
        warnings.warn(
            "BECMessage.loads() is deprecated and should not be used anymore. When calling Connector methods, it can be omitted. When a message needs to be deserialized call the appropriate function from bec_lib.serialization",
            FutureWarning,
        )
        return self

    def dumps(self):
        warnings.warn(
            "BECMessage.dumps() is deprecated and should not be used anymore. When calling Connector methods, it can be omitted. When a message needs to be serialized call the appropriate function from bec_lib.serialization",
            FutureWarning,
        )
        return self


class BundleMessage(BECMessage):
    """Message type to send a bundle of BECMessages.

    Used to bundle together various messages, i.e. used to emit data in the scan bundler.

    Args:
        messages (list): List of BECMessage objects that are bundled together
        metadata (dict, optional): Additional metadata to describe the scan

    Examples:
        >>> BundleMessage(messages=[ScanQueueMessage(...), ScanStatusMessage(...)], metadata = {...})

    """

    msg_type: ClassVar[str] = "bundle_message"
    messages: list = Field(default_factory=list[BECMessage])

    def append(self, msg: BECMessage):
        """Append a new BECMessage to the bundle"""
        if not isinstance(msg, BECMessage):
            raise AttributeError(f"Cannot append message of type {msg.__class__.__name__}")
        # pylint: disable=no-member
        self.messages.append(msg)

    def __len__(self):
        return len(self.messages)

    def __iter__(self):
        # pylint: disable=not-an-iterable
        yield from self.messages


class ScanQueueMessage(BECMessage):
    """Message type for sending scan requests to the scan queue

    Sent by the API server / user to the scan_queue topic. It will be consumed by the scan server.
        Args:
            scan_type (str): one of the registered scan types; either rpc calls or scan types defined in the scan server
            parameter (dict): required parameters for the given scan_stype
            queue (str): either "primary" or "interception"
            metadata (dict, optional): additional metadata to describe the scan
        Examples:
            >>> ScanQueueMessage(scan_type="dscan", parameter={"motor1": "samx", "from_m1:": -5, "to_m1": 5, "steps_m1": 10, "motor2": "samy", "from_m2": -5, "to_m2": 5, "steps_m2": 10, "exp_time": 0.1})
    """

    msg_type: ClassVar[str] = "scan_queue_message"
    scan_type: str
    parameter: dict
    queue: str = Field(default="primary")


class ScanQueueHistoryMessage(BECMessage):
    """Sent after removal from the active queue. Contains information about the scan.

    Called by the ScanWorker after processing the QueueInstructionItem. It can be checked by any service.

    Args:
        status (str): Current scan status
        queue_id (str): Unique queue ID
        info (dict): Dictionary containing additional information about the scan
        queue (str): Defaults to "primary" queue. Information about the queue the scan was in.
        metadata (dict, optional): Additional metadata to describe the scan

    Examples:
        >>> ScanQueueHistoryMessage(status="open", queue_id="1234", info={"positions": {"samx": 0.5, "samy": 0.5}})
    """

    msg_type: ClassVar[str] = "queue_history"
    status: str
    queue_id: str
    info: dict
    queue: str = Field(default="primary")


class ScanStatusMessage(BECMessage):
    """Message type for sending scan status updates.

    Args:
        scan_id (str): Unique scan ID
        status (Literal["open", "paused", "aborted", "halted", "closed"]) : Current scan status
        info (dict): Dictionary containing additional information about the scan
        timestamp (float, optional): Timestamp of the scan status update. If None, the current time is used.
        metadata (dict, optional): Additional metadata to describe and identify the scan.

    Examples:
        >>> ScanStatusMessage(scan_id="1234", status="open", info={"positions": {"samx": 0.5, "samy": 0.5}})
    """

    msg_type: ClassVar[str] = "scan_status"
    scan_id: str | None
    status: Literal["open", "paused", "aborted", "halted", "closed"]
    info: dict
    timestamp: float = Field(default_factory=time.time)

    def __str__(self):
        content = deepcopy(self.__dict__)
        if content["info"].get("positions"):
            content["info"]["positions"] = "..."
        return f"{self.__class__.__name__}({content, self.metadata}))"


class ScanQueueModificationMessage(BECMessage):
    """Message type for sending scan queue modifications

    Args:
        scan_id (str): Unique scan ID
        action (str): One of the actions defined in ACTIONS: ("pause", "deferred_pause", "continue", "abort", "clear", "restart", "halt", "resume")
        parameter (dict): Additional parameters for the action
        queue (str): Defaults to "primary" queue. The name of the queue that receives the modification.
        metadata (dict, optional): Additional metadata to describe and identify the scan.

    Examples:
        >>> ScanQueueModificationMessage(scan_id=scan_id, action="abort", parameter={})
    """

    msg_type: ClassVar[str] = "scan_queue_modification"
    scan_id: str | list[str] | None
    action: Literal[
        "pause", "deferred_pause", "continue", "abort", "clear", "restart", "halt", "resume"
    ]
    parameter: dict
    queue: str = Field(default="primary")


class ScanQueueStatusMessage(BECMessage):
    """Message type for sending scan queue status updates

    Args:
        queue (dict): Dictionary containing the current queue status. Must contain a "primary" key.
        metadata (dict, optional): Additional metadata to describe and identify the ScanQueueStatus.

    Examples:
        >>> ScanQueueStatusMessage(queue={"primary": {}}, metadata={"RID": "1234"})
    """

    msg_type: ClassVar[str] = "scan_queue_status"
    queue: dict

    @field_validator("queue")
    @classmethod
    def check_queue(cls, v):
        """Validate the queue"""
        if not isinstance(v, dict):
            raise ValueError(f"Invalid queue {v}. Must be a dictionary")
        if "primary" not in v:
            raise ValueError(f"Invalid queue {v}. Must contain a 'primary' key")
        return v


class ClientInfoMessage(BECMessage):
    """Message type for sending information to the client
    Args:
        message (str): message to the client
        show_asap (bool, optional): True if the message should be shown immediately. Defaults to True
        # Note: The option show_asap = True/False is temporary disabled until a decision is made on how to handle it. TODO #286
        RID (str, optional): Request ID forwarded from the service, if available will be used to filter on the client site. Defaults to None.
        source (str, Literal[
            "bec_ipython_client",
            "scan_server",
            "device_server",
            "scan_bundler",
            "file_writer",
            "scihub",
            "dap",
            None]
            : Source of the message. Defaults to None.
        scope (str, optional): Scope of the message; Defaults to None. One can follow
                               a pattern to filter afterwards for specific client info; e.g. "scan", "rotation"
        severity (int, optional): severity level of the message (0: INFO, 1: WARNING, 2: ERROR); Defaults to 0
    """

    msg_type: ClassVar[str] = "client_info"
    message: str
    show_asap: bool = Field(default=True)
    RID: str | None = Field(default=None)
    source: Literal[
        "bec_ipython_client",
        "scan_server",
        "device_server",
        "scan_bundler",
        "file_writer",
        "scihub",
        "dap",
        None,
    ] = Field(default=None)
    scope: str | None = Field(default=None)
    severity: int = Field(
        default=0
    )  # TODO add enum for severity levels INFO = 0, WARNING = 1, ERROR = 2


class RequestResponseMessage(BECMessage):
    """Message type for sending back decisions on the acceptance of requests

    Args:
        accepted (bool): True if the request was accepted
        message (str, dict, optional): String or dictionary describing the decision, e.g. "Invalid request"
        metadata (dict, optional): Additional metadata, defaults to None.

    Examples:
        >>> RequestResponseMessage(accepted=True, message="Request accepted")
    """

    msg_type: ClassVar[str] = "request_response"
    accepted: bool
    message: str | dict | None = Field(default=None)


class DeviceInstructionMessage(BECMessage):
    """Message type for sending device instructions to the device server

    Args:
        device (str, list[str], None): Device name, list of device names or None
        action (Literal[ "rpc",
                        "set",
                        "read",
                        "kickoff",
                        "complete",
                        "trigger",
                        "stage",
                        "unstage",
                        "pre_scan",
                        "wait",
                        "scan_report_instruction",
                        "open_scan",
                        "baseline_reading",
                        "close_scan",
                        "open_scan_def",
                        "close_scan_def",
                        "publish_data_as_read",
                        "close_scan_group",
                        ]) : Device action, note rpc calls can run any method of the device. The function name needs to be specified in parameters['func']
        parameter (dict): Parameters required for the device action
        metadata (dict, optional): Metadata to describe the conditions of the device instruction

    Examples:
        >>> DeviceInstructionMessage(device="samx", action="stage", parameter={})
    """

    msg_type: ClassVar[str] = "device_instruction"
    device: str | list[str] | None
    action: Literal[
        "rpc",
        "set",
        "read",
        "kickoff",
        "complete",
        "trigger",
        "stage",
        "unstage",
        "pre_scan",
        "wait",
        "scan_report_instruction",
        "open_scan",
        "baseline_reading",
        "close_scan",
        "open_scan_def",
        "close_scan_def",
        "publish_data_as_read",
        "close_scan_group",
    ]
    parameter: dict


class DeviceMessage(BECMessage):
    """Message type for sending device readings from the device server

    Args:
        signals (dict): Dictionary containing the device signals and their values
        metadata (dict, optional): Metadata to describe the conditions of the device reading

    Examples:
        >>> BECMessage.DeviceMessage(signals={'samx': {'value': 14.999033949016491, 'timestamp': 1686385306.0265112}, 'samx_setpoint': {'value': 15.0, 'timestamp': 1686385306.016806}, 'samx_motor_is_moving': {'value': 0, 'timestamp': 1686385306.026888}}}, metadata={'stream': 'primary', 'DIID': 353, 'RID': 'd3471acc-309d-43b7-8ff8-f986c3fdecf1', 'point_id': 49, 'scan_id': '8e234698-358e-402d-a272-73e168a72f66', 'queue_id': '7a232746-6c90-44f5-81f5-74ab0ea22d4a'})
    """

    msg_type: ClassVar[str] = "device_message"
    signals: dict[str, dict[Literal["value", "timestamp"], Any]]


class DeviceRPCMessage(BECMessage):
    """Message type for sending device RPC return values from the device server

    Args:
        device (str): Device name.
        return_val (Any): Return value of the RPC call.
        out (str or dict): Output of the RPC call.
        success (bool, optional): True if the RPC call was successful. Defaults to True.
        metadata (dict, optional): Additional metadata.

    """

    msg_type: ClassVar[str] = "device_rpc_message"
    device: str
    return_val: Any
    out: str | dict
    success: bool = Field(default=True)


class DeviceStatusMessage(BECMessage):
    """Message type for sending device status updates from the device server

    Args:
        device (str): Device name.
        status (int): Device status.
        metadata (dict, optional): Additional metadata.

    """

    msg_type: ClassVar[str] = "device_status_message"
    device: str
    status: int


class DeviceReqStatusMessage(BECMessage):
    """Message type for sending device request status updates from the device server

    Args:
        device (str): Device name.
        success (bool): True if the request was successful.
        metadata (dict, optional): Additional metadata.

    """

    msg_type: ClassVar[str] = "device_req_status_message"
    device: str
    success: bool


class DeviceInfoMessage(BECMessage):
    """Message type for sending device info updates from the device server

    Args:
        device (str): Device name.
        info (dict): Device info as a dictionary.
        metadata (dict, optional): Additional metadata.

    """

    msg_type: ClassVar[str] = "device_info_message"
    device: str
    info: dict


class DeviceMonitorMessage(BECMessage):
    """Message type for sending device monitor updates from the device server.

    The message is send from the device_server to monitor data coming from larger detector.

    Args:
        device (str): Device name.
        data (np.ndarray): Numpy array data from the monitor
        metadata (dict, optional): Additional metadata.

    """

    msg_type: ClassVar[str] = "device_monitor_message"
    device: str
    data: np.ndarray

    class Config:
        arbitrary_types_allowed = True


class ScanMessage(BECMessage):
    """Message type for sending scan segment data from the scan bundler

    Args:
        point_id (int): Point ID from the scan segment.
        scan_id (str): Scan ID.
        data (dict): Scan segment data.
        metadata (dict, optional): Additional metadata.

    """

    msg_type: ClassVar[str] = "scan_message"
    point_id: int
    scan_id: str
    data: dict


class ScanBaselineMessage(BECMessage):
    """Message type for sending scan baseline data from the scan bundler

    Args:
        scan_id (str): Scan ID.
        data (dict): Scan baseline data.
        metadata (dict, optional): Additional metadata.

    """

    msg_type: ClassVar[str] = "scan_baseline_message"
    scan_id: str
    data: dict


class DeviceConfigMessage(BECMessage):
    """Message type for sending device config updates

    Args:
        action (Literal['add', 'set', 'update', 'reload', or 'remove'] : Update of the device config.
        config (dict, or None): Device config (add, set, update) or None (reload).
        metadata (dict, optional): Additional metadata.

    """

    msg_type: ClassVar[str] = "device_config_message"
    action: Literal["add", "set", "update", "reload", "remove"] = Field(
        default=None, validate_default=True
    )
    config: dict | None = Field(default=None)

    @model_validator(mode="after")
    @classmethod
    def check_config(cls, values):
        """Validate the config"""
        if values.action in ["add", "set", "update"] and not values.config:
            raise ValueError(f"Invalid config {values.config}. Must be a dictionary")
        return values


class LogMessage(BECMessage):
    """Log message

    Args:
        log_type (Literal["trace", "debug", "info", "success", "warning", "error", "critical", "console_log"]) : Log type.
        log_msg (dict or str): Log message.
        metadata (dict, optional): Additional metadata.

    """

    msg_type: ClassVar[str] = "log_message"
    log_type: Literal[
        "trace", "debug", "info", "success", "warning", "error", "critical", "console_log"
    ]
    log_msg: dict | str


class AlarmMessage(BECMessage):
    """Alarm message

    Args:
        severity (Alarms, Literal[0,1,2]): Severity level (0-2). ALARMS.WARNING = 0, ALARMS.MINOR = 1, ALARMS.MAJOR = 2
        alarm_type (str): Type of alarm.
        source (dict): Source of the problem.
        msg (str): Problem description.
        metadata (dict, optional): Additional metadata.

    """

    msg_type: ClassVar[str] = "alarm_message"
    severity: int  # TODO change once enums moved to a separate class
    alarm_type: str
    source: dict
    msg: str


class StatusMessage(BECMessage):
    """Status message

    Args:
        name (str): Name of the status.
        status (BECStatus): Value of the BECStatus enum (RUNNING = 2,  BUSY = 1, IDLE = 0, ERROR = -1).
        info (dict): Status info.
        metadata (dict, optional): Additional metadata.

    """

    msg_type: ClassVar[str] = "status_message"
    name: str
    status: BECStatus
    info: dict


class FileMessage(BECMessage):
    """File message to inform about the status of a file writing operation

    Args:
        file_path (str): Path to the file.
        done (bool, optional): True if the file writing operation is done. Defaults to True.
        successful (bool, optional): True if the file writing operation was successful. Defaults to True.
        metadata (dict, optional): Additional metadata. Defaults to None.

    """

    msg_type: ClassVar[str] = "file_message"
    file_path: str
    done: bool = Field(default=True)
    successful: bool = Field(default=True)


class FileContentMessage(BECMessage):
    """File content message to inform about the content of a file

    Args:
        file_path (str): Path to the file.
        data (str): Content of the file.
        metadata (dict, optional): Status metadata. Defaults to None.

    """

    msg_type: ClassVar[str] = "file_content_message"
    file_path: str
    data: dict


class VariableMessage(BECMessage):
    """Message to inform about a global variable

    Args:
        value (Any): Variable value, can be of any type.
        metadata (dict, optional): Additional metadata.

    """

    msg_type: ClassVar[str] = "var_message"
    value: Any


class ObserverMessage(BECMessage):
    """Message for observer updates

    Args:
        observer (list[dict]): List of observer descriptions (dictionaries).
        metadata (dict, optional): Additional metadata.

    """

    msg_type: ClassVar[str] = "observer_message"
    observer: list[dict]


class ServiceMetricMessage(BECMessage):
    """Message for service metrics

    Args:
        name (str): Name of the service.
        metrics (dict): Dictionary with service metrics.
        metadata (dict, optional): Additional metadata.

    """

    msg_type: ClassVar[str] = "service_metric_message"
    name: str
    metrics: dict


class ProcessedDataMessage(BECMessage):
    """Message for processed data

    Args:
        data (dict, list[dict]): Dictionary with processed data or list of dictionaries with processed data.
        metadata (dict, optional): Metadata. Defaults to None.
    """

    msg_type: ClassVar[str] = "processed_data_message"
    data: dict | list[dict]


class DAPConfigMessage(BECMessage):
    """Message for DAP configuration

    Args:
        config (dict): DAP configuration dictionary
        metadata (dict, optional): Metadata. Defaults to None.
    """

    msg_type: ClassVar[str] = "dap_config_message"
    config: dict


class DAPRequestMessage(BECMessage):
    """Message for DAP requests

    Args:
        dap_cls (str): DAP class name
        dap_type (Literal["continuous", "on_demand"]) : Different types of DAP modes
        config (dict): DAP configuration
        metadata (dict, optional): Metadata. Defaults to None.
    """

    msg_type: ClassVar[str] = "dap_request_message"
    dap_cls: str
    dap_type: Literal["continuous", "on_demand"]
    config: dict


class DAPResponseMessage(BECMessage):
    """Message for DAP responses

    Args:
        success (bool): True if the request was successful
        data (tuple, optional): DAP data (tuple of data (dict) and metadata). Defaults to ({} , None).
        error (str, optional): DAP error. Defaults to None.
        dap_request (BECMessage, None): DAP request. Defaults to None.
        metadata (dict, optional): Metadata. Defaults to None.
    """

    msg_type: ClassVar[str] = "dap_response_message"
    success: bool
    data: tuple | None = Field(default_factory=lambda: ({}, None))
    error: str | None = None
    dap_request: BECMessage | None = Field(default=None)


class AvailableResourceMessage(BECMessage):
    """Message for available resources such as scans, data processing plugins etc

    Args:
        resource (dict, list[dict]): Resource description
        metadata (dict, optional): Metadata. Defaults to None.
    """

    msg_type: ClassVar[str] = "available_resource_message"
    resource: dict | list[dict]


class ProgressMessage(BECMessage):
    """Message for communicating the progress of a long running task

    Args:
        value (float): Current progress value
        max_value (float): Maximum progress value
        done (bool): True if the task is done
        metadata (dict, optional): Metadata. Defaults to None.
    """

    msg_type: ClassVar[str] = "progress_message"
    value: float
    max_value: float
    done: bool


class GUIConfigMessage(BECMessage):
    """Message for GUI configuration

    Args:
        config (dict): GUI configuration, check widgets for more details
        metadata (dict, optional): Metadata. Defaults to None.
    """

    msg_type: ClassVar[str] = "gui_config_message"
    config: dict


class GUIDataMessage(BECMessage):
    """Message for GUI data, i.e. update for DAP processes or scans

    Args:
        data (dict): GUI data
        metadata (dict, optional): Metadata. Defaults to None.
    """

    msg_type: ClassVar[str] = "gui_data_message"
    data: dict


class GUIInstructionMessage(BECMessage):
    """Message for GUI instructions

    Args:
        action (str): Instruction to be executed by the GUI
        metadata (dict, optional): Metadata. Defaults to None.
    """

    msg_type: ClassVar[str] = "gui_instruction_message"
    action: str
    parameter: dict


class ServiceResponseMessage(BECMessage):
    """Message for service responses

    Args:
        response (dict): Service response
        metadata (dict, optional): Metadata. Defaults to None.
    """

    msg_type: ClassVar[str] = "service_response_message"
    response: dict


class CredentialsMessage(BECMessage):
    """Message for credentials

    Args:
        credentials (dict): Credentials
        metadata (dict, optional): Metadata. Defaults to None.
    """

    msg_type: ClassVar[str] = "credentials_message"
    credentials: dict



import datetime
import os
from unittest import mock

import h5py
import numpy as np
import pytest
from test_file_writer_manager import file_writer_manager_mock

from bec_lib.tests.fixtures import dm_with_devices
from bec_server import file_writer
from bec_server.file_writer import NexusFileWriter
from bec_server.file_writer.file_writer import HDF5Storage
from bec_server.file_writer.file_writer_manager import ScanStorage
from bec_server.file_writer_plugins.cSAXS import NeXus_format as cSAXS_Nexus_format

dir_path = os.path.dirname(file_writer.__file__)


@pytest.fixture
def file_writer_manager_mock_with_dm(file_writer_manager_mock, dm_with_devices):
    file_writer = file_writer_manager_mock
    file_writer.device_manager = dm_with_devices
    yield file_writer


def test_csaxs_nexus_format(file_writer_manager_mock_with_dm):
    file_manager = file_writer_manager_mock_with_dm
    writer_storage = cSAXS_Nexus_format(
        storage=HDF5Storage(),
        data={"samx": {"samx": {"value": [0, 1, 2]}}, "mokev": {"mokev": {"value": 12.456}}},
        file_references={},
        device_manager=file_manager.device_manager,
    )
    assert writer_storage._storage["entry"].attrs["definition"] == "NXsas"
    assert writer_storage._storage["entry"]._storage["sample"]._storage["x_translation"]._data == [
        0,
        1,
        2,
    ]


def test_nexus_file_writer(file_writer_manager_mock_with_dm):
    file_manager = file_writer_manager_mock_with_dm
    file_writer = NexusFileWriter(file_manager)
    with mock.patch.object(
        file_writer,
        "_create_device_data_storage",
        return_value={
            "samx": [
                {"samx": {"value": 0}},
                {"samx": {"value": 1}},
                {"samx": {"value": 2}},
                {"samx": {"value": 3}},
                {"samx": {"value": 4}},
            ]
        },
    ):
        file_writer.write("./test.h5", ScanStorage(2, "scan_id-string"))
    with h5py.File("./test.h5", "r") as test_file:
        assert list(test_file) == ["entry"]
        assert list(test_file["entry"]) == ["collection", "control", "instrument", "sample"]
        assert np.allclose(test_file["entry/collection/bec/samx/samx/value"][...], [0, 1, 2, 3, 4])
        # assert list(test_file["entry"]["sample"]) == ["x_translation"]
        # assert test_file["entry"]["sample"].attrs["NX_class"] == "NXsample"
        # assert test_file["entry"]["sample"]["x_translation"].attrs["units"] == "mm"
        # assert all(np.asarray(test_file["entry"]["sample"]["x_translation"]) == [0, 1, 2])


def test_create_device_data_storage(file_writer_manager_mock_with_dm):
    file_manager = file_writer_manager_mock_with_dm
    file_writer = NexusFileWriter(file_manager)
    storage = ScanStorage("2", "scan_id-string")
    storage.num_points = 2
    storage.scan_segments = {
        0: {"samx": {"samx": {"value": 0.1}}, "samy": {"samy": {"value": 1.1}}},
        1: {"samx": {"samx": {"value": 0.2}}, "samy": {"samy": {"value": 1.2}}},
    }
    storage.baseline = {}
    device_storage = file_writer._create_device_data_storage(storage)
    assert len(device_storage.keys()) == 2
    assert len(device_storage["samx"]) == 2
    assert device_storage["samx"][0]["samx"]["value"] == 0.1
    assert device_storage["samx"][1]["samx"]["value"] == 0.2


@pytest.mark.parametrize(
    "segments,baseline,metadata",
    [
        (
            {
                0: {
                    "samx": {"samx": {"value": 0.11}, "samx_setpoint": {"value": 0.1}},
                    "samy": {"samy": {"value": 1.1}},
                },
                1: {
                    "samx": {"samx": {"value": 0.21}, "samx_setpoint": {"value": 0.2}},
                    "samy": {"samy": {"value": 1.2}},
                },
            },
            {
                "eyefoc": {
                    "eyefoc": {"value": 0, "timestamp": 1679226971.564248},
                    "eyefoc_setpoint": {"value": 0, "timestamp": 1679226971.564235},
                    "eyefoc_motor_is_moving": {"value": 0, "timestamp": 1679226971.564249},
                },
                "field": {
                    "field_x": {"value": 0, "timestamp": 1679226971.579148},
                    "field_x_setpoint": {"value": 0, "timestamp": 1679226971.579145},
                    "field_x_motor_is_moving": {"value": 0, "timestamp": 1679226971.579148},
                    "field_y": {"value": 0, "timestamp": 1679226971.5799649},
                    "field_y_setpoint": {"value": 0, "timestamp": 1679226971.579962},
                    "field_y_motor_is_moving": {"value": 0, "timestamp": 1679226971.579966},
                    "field_z_zsub": {"value": 0, "timestamp": 1679226971.58087},
                    "field_z_zsub_setpoint": {"value": 0, "timestamp": 1679226971.580867},
                    "field_z_zsub_motor_is_moving": {"value": 0, "timestamp": 1679226971.58087},
                },
            },
            {
                "RID": "5ee455b8-d0ef-452d-b54a-e7cea5cea19e",
                "scan_id": "a9fb36e4-3f38-486c-8434-c8eca19472ba",
                "queue_id": "14463a5b-1c65-4888-8f87-4808c90a241f",
                "primary": ["samx"],
                "num_points": 2,
                "positions": [[-100], [100]],
                "scan_name": "monitor_scan",
                "scan_type": "fly",
                "scan_number": 88,
                "dataset_number": 88,
                "exp_time": 0.1,
                "scan_report_hint": "table",
                "scan_report_devices": ["samx"],
                "scan_msgs": [
                    "ScanQueueMessage(({'scan_type': 'monitor_scan', 'parameter': {'args': {'samx':"
                    " [-100, 100]}, 'kwargs': {'relative': False}}, 'queue': 'primary'}, {'RID':"
                    " '5ee455b8-d0ef-452d-b54a-e7cea5cea19e'})))"
                ],
            },
        )
    ],
)
def test_write_data_storage(segments, baseline, metadata, file_writer_manager_mock_with_dm):
    file_manager = file_writer_manager_mock_with_dm
    file_writer = NexusFileWriter(file_manager)
    storage = ScanStorage("2", "scan_id-string")
    storage.num_points = 2
    storage.scan_segments = segments
    storage.baseline = baseline
    storage.metadata = metadata
    storage.start_time = 1679226971.564235
    storage.end_time = 1679226971.580867

    file_writer.write("./test.h5", storage)

    # open file and check that time stamps are correct
    with h5py.File("./test.h5", "r") as test_file:
        assert (
            test_file["entry"].attrs["start_time"]
            == datetime.datetime.fromtimestamp(1679226971.564235).isoformat()
        )

        assert (
            test_file["entry"].attrs["end_time"]
            == datetime.datetime.fromtimestamp(1679226971.580867).isoformat()
        )



from bec_lib.device import DeviceBase
from bec_lib.scan_report import ScanReport

# pylint:disable=undefined-variable
# pylint: disable=too-many-arguments


def dscan(
    motor1: DeviceBase, m1_from: float, m1_to: float, steps: int, exp_time: float, **kwargs
) -> ScanReport:
    """Relative line scan with one device.

    Args:
        motor1 (Device): Device that should be scanned.
        m1_from (float): Start position relative to the current position.
        m1_to (float): End position relative to the current position.
        steps (int): Number of steps.
        exp_time (float): Exposure time.

    Returns:
        ScanReport: Status object.

    Examples:
        >>> dscan(dev.motor1, -5, 5, 10, 0.1)
    """
    return scans.line_scan(
        motor1, m1_from, m1_to, steps=steps, exp_time=exp_time, relative=True, **kwargs
    )


def d2scan(
    motor1: DeviceBase,
    m1_from: float,
    m1_to: float,
    motor2: DeviceBase,
    m2_from: float,
    m2_to: float,
    steps: int,
    exp_time: float,
    **kwargs
) -> ScanReport:
    """Relative line scan with two devices.

    Args:
        motor1 (Device): First device that should be scanned.
        m1_from (float): Start position of the first device relative to its current position.
        m1_to (float): End position of the first device relative to its current position.
        motor2 (Device): Second device that should be scanned.
        m2_from (float): Start position of the second device relative to its current position.
        m2_to (float): End position of the second device relative to its current position.
        steps (int): Number of steps.
        exp_time (float): Exposure time

    Returns:
        ScanReport: Status object.

    Examples:
        >>> d2scan(dev.motor1, -5, 5, dev.motor2, -8, 8, 10, 0.1)
    """
    return scans.line_scan(
        motor1,
        m1_from,
        m1_to,
        motor2,
        m2_from,
        m2_to,
        steps=steps,
        exp_time=exp_time,
        relative=True,
        **kwargs
    )


def ascan(motor1, m1_from, m1_to, steps, exp_time, **kwargs) -> ScanReport:
    """Absolute line scan with one device.

    Args:
        motor1 (Device): Device that should be scanned.
        m1_from (float): Start position.
        m1_to (float): End position.
        steps (int): Number of steps.
        exp_time (float): Exposure time.

    Returns:
        ScanReport: Status object.

    Examples:
        >>> ascan(dev.motor1, -5, 5, 10, 0.1)
    """
    return scans.line_scan(
        motor1, m1_from, m1_to, steps=steps, exp_time=exp_time, relative=False, **kwargs
    )


def a2scan(motor1, m1_from, m1_to, motor2, m2_from, m2_to, steps, exp_time, **kwargs) -> ScanReport:
    """Absolute line scan with two devices.

    Args:
        motor1 (Device): First device that should be scanned.
        m1_from (float): Start position of the first device.
        m1_to (float): End position of the first device.
        motor2 (Device): Second device that should be scanned.
        m2_from (float): Start position of the second device.
        m2_to (float): End position of the second device.
        steps (int): Number of steps.
        exp_time (float): Exposure time

    Returns:
        ScanReport: Status object.

    Examples:
        >>> a2scan(dev.motor1, -5, 5, dev.motor2, -8, 8, 10, 0.1)
    """
    return scans.line_scan(
        motor1,
        m1_from,
        m1_to,
        motor2,
        m2_from,
        m2_to,
        steps=steps,
        exp_time=exp_time,
        relative=False,
        **kwargs
    )


def dmesh(
    motor1, m1_from, m1_to, m1_steps, motor2, m2_from, m2_to, m2_steps, exp_time, **kwargs
) -> ScanReport:
    """Relative mesh scan (grid scan) with two devices.

    Args:
        motor1 (Device): First device that should be scanned.
        m1_from (float): Start position of the first device relative to its current position.
        m1_to (float): End position of the first device relative to its current position.
        m1_steps (int): Number of steps for motor1.
        motor2 (Device): Second device that should be scanned.
        m2_from (float): Start position of the second device relative to its current position.
        m2_to (float): End position of the second device relative to its current position.
        m2_steps (int): Number of steps for motor2.
        exp_time (float): Exposure time

    Returns:
        ScanReport: Status object.

    Examples:
        >>> dmesh(dev.motor1, -5, 5, 10, dev.motor2, -8, 8, 10, 0.1)
    """
    return scans.grid_scan(
        motor1,
        m1_from,
        m1_to,
        m1_steps,
        motor2,
        m2_from,
        m2_to,
        m2_steps,
        exp_time=exp_time,
        relative=True,
    )


def amesh(
    motor1, m1_from, m1_to, m1_steps, motor2, m2_from, m2_to, m2_steps, exp_time, **kwargs
) -> ScanReport:
    """Absolute mesh scan (grid scan) with two devices.

    Args:
        motor1 (Device): First device that should be scanned.
        m1_from (float): Start position of the first device.
        m1_to (float): End position of the first device.
        m1_steps (int): Number of steps for motor1.
        motor2 (Device): Second device that should be scanned.
        m2_from (float): Start position of the second device.
        m2_to (float): End position of the second device.
        m2_steps (int): Number of steps for motor2.
        exp_time (float): Exposure time

    Returns:
        ScanReport: Status object.

    Examples:
        >>> amesh(dev.motor1, -5, 5, 10, dev.motor2, -8, 8, 10, 0.1)
    """
    return scans.grid_scan(
        motor1,
        m1_from,
        m1_to,
        m1_steps,
        motor2,
        m2_from,
        m2_to,
        m2_steps,
        exp_time=exp_time,
        relative=False,
    )



# pylint: disable=missing-function-docstring
import threading
import time
from contextlib import redirect_stdout
from io import StringIO
from unittest import mock

import numpy as np
import pytest

from bec_ipython_client.callbacks.live_table import LiveUpdatesTable, sort_devices
from bec_ipython_client.callbacks.utils import ScanRequestMixin
from bec_lib import messages


@pytest.fixture
def client_with_grid_scan(bec_client_mock):
    client = bec_client_mock
    request_msg = messages.ScanQueueMessage(
        scan_type="grid_scan",
        parameter={"args": {"samx": (-5, 5, 3)}, "kwargs": {}},
        queue="primary",
        metadata={"RID": "something"},
    )
    yield client, request_msg


@pytest.mark.timeout(20)
def test_scan_request_mixin(client_with_grid_scan):
    client, request_msg = client_with_grid_scan
    response_msg = messages.RequestResponseMessage(
        accepted=True, message={"msg": ""}, metadata={"RID": "something"}
    )
    request_mixin = ScanRequestMixin(client, "something")

    def update_with_response(request_msg):
        time.sleep(1)
        client.queue.request_storage.update_with_response(response_msg)

    client.queue.request_storage.update_with_request(request_msg)
    update_thread = threading.Thread(target=update_with_response, args=(response_msg,))
    update_thread.start()
    with mock.patch.object(client.queue.queue_storage, "find_queue_item_by_requestID"):
        request_mixin.wait()
    update_thread.join()


def test_sort_devices():
    devices = sort_devices(["samx", "bpm4i", "samy", "bpm4s"], ["samx", "samy"])
    assert devices == ["samx", "samy", "bpm4i", "bpm4s"]


@pytest.mark.parametrize(
    "request_msg,scan_report_devices",
    [
        (
            messages.ScanQueueMessage(
                scan_type="grid_scan",
                parameter={"args": {"samx": (-5, 5, 3)}, "kwargs": {}},
                queue="primary",
                metadata={"RID": "something"},
            ),
            ["samx"],
        ),
        (
            messages.ScanQueueMessage(
                scan_type="round_scan",
                parameter={"args": {"samx": ["samy", 0, 25, 5, 3]}},
                queue="primary",
                metadata={"RID": "something"},
            ),
            ["samx", "samy"],
        ),
    ],
)
def test_get_devices_from_scan_data(bec_client_mock, request_msg, scan_report_devices):
    client = bec_client_mock
    client.start()
    data = messages.ScanMessage(
        point_id=0, scan_id="", data={}, metadata={"scan_report_devices": scan_report_devices}
    )
    live_update = LiveUpdatesTable(client, {"table_wait": 10}, request_msg)
    devices = live_update.get_devices_from_scan_data(data)
    assert devices[0 : len(scan_report_devices)] == scan_report_devices


@pytest.mark.timeout(20)
def test_wait_for_request_acceptance(client_with_grid_scan):
    client, request_msg = client_with_grid_scan
    response_msg = messages.RequestResponseMessage(
        accepted=True, message={"msg": ""}, metadata={"RID": "something"}
    )
    client.queue.request_storage.update_with_request(request_msg)
    client.queue.request_storage.update_with_response(response_msg)
    live_update = LiveUpdatesTable(client, {"table_wait": 10}, request_msg)
    with mock.patch.object(client.queue.queue_storage, "find_queue_item_by_requestID"):
        live_update.wait_for_request_acceptance()


class ScanItemMock:
    def __init__(self, data):
        self.data = data
        self.metadata = {}


def test_print_table_data(client_with_grid_scan):
    client, request_msg = client_with_grid_scan
    response_msg = messages.RequestResponseMessage(
        accepted=True, message={"msg": ""}, metadata={"RID": "something"}
    )
    client.queue.request_storage.update_with_request(request_msg)
    client.queue.request_storage.update_with_response(response_msg)
    live_update = LiveUpdatesTable(client, {"table_wait": 10}, request_msg)
    live_update.point_data = messages.ScanMessage(
        point_id=0,
        scan_id="",
        data={"samx": {"samx": {"value": 0}}},
        metadata={"scan_report_devices": ["samx"], "scan_type": "step"},
    )
    live_update.scan_item = ScanItemMock(data=[live_update.point_data])
    with mock.patch.object(live_update, "_print_client_msgs_asap") as mock_client_msgs:
        live_update.print_table_data()
        assert mock_client_msgs.called


def test_print_table_data_lamni_flyer(client_with_grid_scan):
    client, request_msg = client_with_grid_scan
    response_msg = messages.RequestResponseMessage(
        accepted=True, message={"msg": ""}, metadata={"RID": "something"}
    )
    client.queue.request_storage.update_with_request(request_msg)
    client.queue.request_storage.update_with_response(response_msg)
    live_update = LiveUpdatesTable(client, {"table_wait": 10}, request_msg)
    live_update.point_data = messages.ScanMessage(
        point_id=0,
        scan_id="",
        data={"lamni_flyer_1": {"value": 0}},
        metadata={"scan_report_devices": ["samx"], "scan_type": "fly"},
    )
    live_update.scan_item = ScanItemMock(data=[live_update.point_data])
    with mock.patch.object(live_update, "_print_client_msgs_asap") as mock_client_msgs:
        live_update.print_table_data()
        assert mock_client_msgs.called


def test_print_table_data_hinted_value(client_with_grid_scan):
    client, request_msg = client_with_grid_scan
    response_msg = messages.RequestResponseMessage(
        accepted=True, message={"msg": ""}, metadata={"RID": "something"}
    )
    client.queue.request_storage.update_with_request(request_msg)
    client.queue.request_storage.update_with_response(response_msg)
    live_update = LiveUpdatesTable(client, {"table_wait": 10}, request_msg)
    client.device_manager.devices["samx"]._info["hints"] = {"fields": ["samx_hint"]}
    client.device_manager.devices["samx"].precision = 3
    live_update.point_data = messages.ScanMessage(
        point_id=0,
        scan_id="",
        data={"samx": {"samx_hint": {"value": 0}}},
        metadata={"scan_report_devices": ["samx"], "scan_type": "fly"},
    )
    live_update.scan_item = ScanItemMock(data=[live_update.point_data])

    with (
        mock.patch.object(live_update, "table") as mocked_table,
        mock.patch.object(live_update, "_print_client_msgs_asap") as mock_client_msgs,
    ):
        live_update.dev_values = (len(live_update._get_header()) - 1) * [0]
        live_update.print_table_data()
        mocked_table.get_row.assert_called_with("0", "0.000")
        assert mock_client_msgs.called


def test_print_table_data_hinted_value_with_precision(client_with_grid_scan):
    client, request_msg = client_with_grid_scan
    response_msg = messages.RequestResponseMessage(
        accepted=True, message={"msg": ""}, metadata={"RID": "something"}
    )
    client.queue.request_storage.update_with_request(request_msg)
    client.queue.request_storage.update_with_response(response_msg)
    live_update = LiveUpdatesTable(client, {"table_wait": 10}, request_msg)
    client.device_manager.devices["samx"]._info["hints"] = {"fields": ["samx_hint"]}
    client.device_manager.devices["samx"].precision = 2
    live_update.point_data = messages.ScanMessage(
        point_id=0,
        scan_id="",
        data={"samx": {"samx_hint": {"value": 0}}},
        metadata={"scan_report_devices": ["samx"], "scan_type": "fly"},
    )
    live_update.scan_item = ScanItemMock(data=[live_update.point_data])

    with (
        mock.patch.object(live_update, "table") as mocked_table,
        mock.patch.object(live_update, "_print_client_msgs_asap") as mock_client_msgs,
    ):
        live_update.dev_values = (len(live_update._get_header()) - 1) * [0]
        live_update.print_table_data()
        mocked_table.get_row.assert_called_with("0", f"{0:.2f}")


@pytest.mark.parametrize(
    "value,expected",
    [
        (0, "0.00"),
        (1, "1.00"),
        (0.000, "0.00"),
        (True, "1.00"),
        (False, "0.00"),
        ("True", "True"),
        ("False", "False"),
        ("0", "0"),
        ("1", "1"),
        ((0, 1), "(0, 1)"),
        ({"value": 0}, "{'value': 0}"),
        (np.array([0, 1]), "[0 1]"),
        ({1, 2}, "{1, 2}"),
    ],
)
def test_print_table_data_variants(client_with_grid_scan, value, expected):
    client, request_msg = client_with_grid_scan
    response_msg = messages.RequestResponseMessage(
        accepted=True, message={"msg": ""}, metadata={"RID": "something"}
    )
    client.queue.request_storage.update_with_request(request_msg)
    client.queue.request_storage.update_with_response(response_msg)
    live_update = LiveUpdatesTable(client, {"table_wait": 10}, request_msg)
    live_update.point_data = messages.ScanMessage(
        point_id=0,
        scan_id="",
        data={"lamni_flyer_1": {"value": value}},
        metadata={"scan_report_devices": ["samx"], "scan_type": "fly"},
    )
    live_update.scan_item = ScanItemMock(data=[live_update.point_data])

    with mock.patch.object(live_update, "_print_client_msgs_asap") as mock_client_msgs:
        live_update.print_table_data()
        with mock.patch.object(live_update, "table") as mocked_table:
            live_update.dev_values = (len(live_update._get_header()) - 1) * [value]
            live_update.print_table_data()
            mocked_table.get_row.assert_called_with("0", expected)


def test_print_client_msgs(client_with_grid_scan):
    client, request_msg = client_with_grid_scan
    response_msg = messages.RequestResponseMessage(
        accepted=True, message={"msg": ""}, metadata={"RID": "something"}
    )
    client.queue.request_storage.update_with_request(request_msg)
    client.queue.request_storage.update_with_response(response_msg)
    client_msg = messages.ClientInfoMessage(
        message="message", RID="something", show_asap=True, source="scan_server"
    )
    client.queue.request_storage.update_with_client_message(client_msg)
    with mock.patch.object(
        client.queue.request_storage.scan_manager.queue_storage, "find_queue_item_by_requestID"
    ):
        live_update = LiveUpdatesTable(client, {"table_wait": 10}, request_msg)
        live_update.wait_for_request_acceptance()
        result = StringIO()
        with redirect_stdout(result):
            live_update._print_client_msgs_asap()
            rtr1 = "Client info (scan_server) : message" + "\n"
            assert result.getvalue() == rtr1
            # second time should not add anything
            live_update._print_client_msgs_asap()
            assert result.getvalue() == rtr1
            # live_update._print_client_msgs_all()
            # rtr2 = (
            #     "------------------------"
            #     + "\n"
            #     + "Summary of client messages"
            #     + "\n"
            #     + "------------------------"
            #     + "\n"
            #     + "Client info (scan_server) : message"
            #     + "\n"
            #     + "------------------------"
            #     + "\n"
            # )
            # assert result.getvalue() == rtr1 + rtr2



import os
import threading
import time as ttime

import numpy as np
from bec_lib import messages
from bec_lib.endpoints import MessageEndpoints
from bec_lib.logger import bec_logger
from ophyd import Component as Cpt
from ophyd import Device, DeviceStatus
from ophyd import DynamicDeviceComponent as Dcpt
from ophyd import Kind, PositionerBase
from ophyd.flyers import FlyerInterface
from ophyd.sim import SynSignal
from ophyd.status import StatusBase
from ophyd.utils import LimitError

from ophyd_devices.sim.sim_data import (
    SimulatedDataCamera,
    SimulatedDataMonitor,
    SimulatedDataWaveform,
    SimulatedPositioner,
)
from ophyd_devices.sim.sim_signals import ReadOnlySignal, SetableSignal
from ophyd_devices.sim.sim_test_devices import DummyController
from ophyd_devices.utils.bec_scaninfo_mixin import BecScaninfoMixin

logger = bec_logger.logger


class DeviceStop(Exception):
    pass


class SimMonitor(Device):
    """
    A simulated device mimic any 1D Axis (position, temperature, beam).

    It's readback is a computed signal, which is configurable by the user and from the command line.
    The corresponding simulation class is sim_cls=SimulatedDataMonitor, more details on defaults within the simulation class.

    >>> monitor = SimMonitor(name="monitor")

    Parameters
    ----------
    name (string)           : Name of the device. This is the only required argmuent, passed on to all signals of the device.
    precision (integer)     : Precision of the readback in digits, written to .describe(). Default is 3 digits.
    sim_init (dict)         : Dictionary to initiate parameters of the simulation, check simulation type defaults for more details.
    parent                  : Parent device, optional, is used internally if this signal/device is part of a larger device.
    kind                    : A member the Kind IntEnum (or equivalent integer), optional. Default is Kind.normal. See Kind for options.
    device_manager          : DeviceManager from BEC, optional . Within startup of simulation, device_manager is passed on automatically.

    """

    USER_ACCESS = ["sim", "registered_proxies"]

    sim_cls = SimulatedDataMonitor
    BIT_DEPTH = np.uint32

    readback = Cpt(ReadOnlySignal, value=BIT_DEPTH(0), kind=Kind.hinted, compute_readback=True)

    SUB_READBACK = "readback"
    _default_sub = SUB_READBACK

    def __init__(
        self,
        name,
        *,
        precision: int = 3,
        sim_init: dict = None,
        parent=None,
        kind=None,
        device_manager=None,
        **kwargs,
    ):
        self.precision = precision
        self.init_sim_params = sim_init
        self.sim = self.sim_cls(parent=self, device_manager=device_manager, **kwargs)
        self._registered_proxies = {}

        super().__init__(name=name, parent=parent, kind=kind, **kwargs)
        self.sim.sim_state[self.name] = self.sim.sim_state.pop(self.readback.name, None)
        self.readback.name = self.name

    @property
    def registered_proxies(self) -> None:
        """Dictionary of registered signal_names and proxies."""
        return self._registered_proxies


class SimCamera(Device):
    """A simulated device mimic any 2D camera.

    It's image is a computed signal, which is configurable by the user and from the command line.
    The corresponding simulation class is sim_cls=SimulatedDataCamera, more details on defaults within the simulation class.

    >>> camera = SimCamera(name="camera")

    Parameters
    ----------
    name (string)           : Name of the device. This is the only required argmuent, passed on to all signals of the device.
    precision (integer)     : Precision of the readback in digits, written to .describe(). Default is 3 digits.
    sim_init (dict)         : Dictionary to initiate parameters of the simulation, check simulation type defaults for more details.
    parent                  : Parent device, optional, is used internally if this signal/device is part of a larger device.
    kind                    : A member the Kind IntEnum (or equivalent integer), optional. Default is Kind.normal. See Kind for options.
    device_manager          : DeviceManager from BEC, optional . Within startup of simulation, device_manager is passed on automatically.

    """

    USER_ACCESS = ["sim", "registered_proxies"]

    sim_cls = SimulatedDataCamera
    SHAPE = (100, 100)
    BIT_DEPTH = np.uint16

    SUB_MONITOR = "monitor"
    _default_sub = SUB_MONITOR

    exp_time = Cpt(SetableSignal, name="exp_time", value=1, kind=Kind.config)
    file_path = Cpt(SetableSignal, name="file_path", value="", kind=Kind.config)
    file_pattern = Cpt(SetableSignal, name="file_pattern", value="", kind=Kind.config)
    frames = Cpt(SetableSignal, name="frames", value=1, kind=Kind.config)
    burst = Cpt(SetableSignal, name="burst", value=1, kind=Kind.config)

    image_shape = Cpt(SetableSignal, name="image_shape", value=SHAPE, kind=Kind.config)
    image = Cpt(
        ReadOnlySignal,
        name="image",
        value=np.empty(SHAPE, dtype=BIT_DEPTH),
        compute_readback=True,
        kind=Kind.omitted,
    )

    def __init__(
        self, name, *, kind=None, parent=None, sim_init: dict = None, device_manager=None, **kwargs
    ):
        self.device_manager = device_manager
        self.init_sim_params = sim_init
        self._registered_proxies = {}
        self.sim = self.sim_cls(parent=self, device_manager=device_manager, **kwargs)

        super().__init__(name=name, parent=parent, kind=kind, **kwargs)
        self._stopped = False
        self._staged = False
        self.scaninfo = None
        self._update_scaninfo()

    @property
    def registered_proxies(self) -> None:
        """Dictionary of registered signal_names and proxies."""
        return self._registered_proxies

    def trigger(self) -> DeviceStatus:
        """Trigger the camera to acquire images.

        This method can be called from BEC during a scan. It will acquire images and send them to BEC.
        Whether the trigger is send from BEC is determined by the softwareTrigger argument in the device config.

        Here, we also run a callback on SUB_MONITOR to send the image data the device_monitor endpoint in BEC.
        """
        status = DeviceStatus(self)

        self.subscribe(status._finished, event_type=self.SUB_ACQ_DONE, run=False)

        def acquire():
            try:
                for _ in range(self.burst.get()):
                    self._run_subs(sub_type=self.SUB_MONITOR, value=self.image.get())
                    if self._stopped:
                        raise DeviceStop
            except DeviceStop:
                pass
            finally:
                self._stopped = False
                self._done_acquiring()

        threading.Thread(target=acquire, daemon=True).start()
        return status

    def _update_scaninfo(self) -> None:
        """Update scaninfo from BecScaninfoMixing
        This depends on device manager and operation/sim_mode
        """
        self.scaninfo = BecScaninfoMixin(self.device_manager)

    def stage(self) -> list[object]:
        """Stage the camera for upcoming scan

        This method is called from BEC in preparation of a scan.
        It receives metadata about the scan from BEC,
        compiles it and prepares the camera for the scan.

        FYI: No data is written to disk in the simulation, but upon each trigger it
        is published to the device_monitor endpoint in REDIS.
        """
        if self._staged:
            return super().stage()
        self.scaninfo.load_scan_metadata()
        self.file_path.set(
            os.path.join(
                self.file_path.get(), self.file_pattern.get().format(self.scaninfo.scan_number)
            )
        )
        self.frames.set(self.scaninfo.num_points * self.scaninfo.frames_per_trigger)
        self.exp_time.set(self.scaninfo.exp_time)
        self.burst.set(self.scaninfo.frames_per_trigger)
        self._stopped = False
        return super().stage()

    def unstage(self) -> list[object]:
        """Unstage the device

        Send reads from all config signals to redis
        """
        if self._stopped is True or not self._staged:
            return super().unstage()

        return super().unstage()

    def stop(self, *, success=False):
        """Stop the device"""
        self._stopped = True
        super().stop(success=success)


class SimWaveform(Device):
    """A simulated device mimic any 1D Waveform detector.

    It's waveform is a computed signal, which is configurable by the user and from the command line.
    The corresponding simulation class is sim_cls=SimulatedDataWaveform, more details on defaults within the simulation class.

    >>> waveform = SimWaveform(name="waveform")

    Parameters
    ----------
    name (string)           : Name of the device. This is the only required argmuent, passed on to all signals of the device.
    precision (integer)     : Precision of the readback in digits, written to .describe(). Default is 3 digits.
    sim_init (dict)         : Dictionary to initiate parameters of the simulation, check simulation type defaults for more details.
    parent                  : Parent device, optional, is used internally if this signal/device is part of a larger device.
    kind                    : A member the Kind IntEnum (or equivalent integer), optional. Default is Kind.normal. See Kind for options.
    device_manager          : DeviceManager from BEC, optional . Within startup of simulation, device_manager is passed on automatically.

    """

    USER_ACCESS = ["sim", "registered_proxies"]

    sim_cls = SimulatedDataWaveform
    SHAPE = (1000,)
    BIT_DEPTH = np.uint16

    SUB_MONITOR = "monitor"
    _default_sub = SUB_MONITOR

    exp_time = Cpt(SetableSignal, name="exp_time", value=1, kind=Kind.config)
    file_path = Cpt(SetableSignal, name="file_path", value="", kind=Kind.config)
    file_pattern = Cpt(SetableSignal, name="file_pattern", value="", kind=Kind.config)
    frames = Cpt(SetableSignal, name="frames", value=1, kind=Kind.config)
    burst = Cpt(SetableSignal, name="burst", value=1, kind=Kind.config)

    waveform_shape = Cpt(SetableSignal, name="waveform_shape", value=SHAPE, kind=Kind.config)
    waveform = Cpt(
        ReadOnlySignal,
        name="waveform",
        value=np.empty(SHAPE, dtype=BIT_DEPTH),
        compute_readback=True,
        kind=Kind.omitted,
    )

    def __init__(
        self, name, *, kind=None, parent=None, sim_init: dict = None, device_manager=None, **kwargs
    ):
        self.device_manager = device_manager
        self.init_sim_params = sim_init
        self._registered_proxies = {}
        self.sim = self.sim_cls(parent=self, device_manager=device_manager, **kwargs)

        super().__init__(name=name, parent=parent, kind=kind, **kwargs)
        self._stopped = False
        self._staged = False
        self.scaninfo = None
        self._update_scaninfo()

    @property
    def registered_proxies(self) -> None:
        """Dictionary of registered signal_names and proxies."""
        return self._registered_proxies

    def trigger(self) -> DeviceStatus:
        """Trigger the camera to acquire images.

        This method can be called from BEC during a scan. It will acquire images and send them to BEC.
        Whether the trigger is send from BEC is determined by the softwareTrigger argument in the device config.

        Here, we also run a callback on SUB_MONITOR to send the image data the device_monitor endpoint in BEC.
        """
        status = DeviceStatus(self)

        self.subscribe(status._finished, event_type=self.SUB_ACQ_DONE, run=False)

        def acquire():
            try:
                for _ in range(self.burst.get()):
                    self._run_subs(sub_type=self.SUB_MONITOR, value=self.waveform.get())
                    if self._stopped:
                        raise DeviceStop
            except DeviceStop:
                pass
            finally:
                self._stopped = False
                self._done_acquiring()

        threading.Thread(target=acquire, daemon=True).start()
        return status

    def _update_scaninfo(self) -> None:
        """Update scaninfo from BecScaninfoMixing
        This depends on device manager and operation/sim_mode
        """
        self.scaninfo = BecScaninfoMixin(self.device_manager)

    def stage(self) -> list[object]:
        """Stage the camera for upcoming scan

        This method is called from BEC in preparation of a scan.
        It receives metadata about the scan from BEC,
        compiles it and prepares the camera for the scan.

        FYI: No data is written to disk in the simulation, but upon each trigger it
        is published to the device_monitor endpoint in REDIS.
        """
        if self._staged:
            return super().stage()
        self.scaninfo.load_scan_metadata()
        self.file_path.set(
            os.path.join(
                self.file_path.get(), self.file_pattern.get().format(self.scaninfo.scan_number)
            )
        )
        self.frames.set(self.scaninfo.num_points * self.scaninfo.frames_per_trigger)
        self.exp_time.set(self.scaninfo.exp_time)
        self.burst.set(self.scaninfo.frames_per_trigger)
        self._stopped = False
        return super().stage()

    def unstage(self) -> list[object]:
        """Unstage the device

        Send reads from all config signals to redis
        """
        if self._stopped is True or not self._staged:
            return super().unstage()

        return super().unstage()

    def stop(self, *, success=False):
        """Stop the device"""
        self._stopped = True
        super().stop(success=success)


class SimPositioner(Device, PositionerBase):
    """
    A simulated device mimicing any 1D Axis device (position, temperature, rotation).

    >>> motor = SimPositioner(name="motor")

    Parameters
    ----------
    name (string)           : Name of the device. This is the only required argmuent, passed on to all signals of the device.\
    Optional parameters:
    ----------
    delay (int)             : If 0, execution of move will be instant. If 1, exectution will depend on motor velocity. Default is 1.
    update_frequency (int)  : Frequency in Hz of the update of the simulated state during a move. Default is 2 Hz.
    precision (integer)     : Precision of the readback in digits, written to .describe(). Default is 3 digits.
    tolerance (float)       : Tolerance of the positioner to accept reaching target positions. Default is 0.5.
    limits (tuple)          : Tuple of the low and high limits of the positioner. Overrides low/high_limit_travel is specified. Default is None.
    parent                  : Parent device, optional, is used internally if this signal/device is part of a larger device.
    kind                    : A member the Kind IntEnum (or equivalent integer), optional. Default is Kind.normal. See Kind for options.
    device_manager          : DeviceManager from BEC, optional . Within startup of simulation, device_manager is passed on automatically.
    sim_init (dict)         : Dictionary to initiate parameters of the simulation, check simulation type defaults for more details.

    """

    # Specify which attributes are accessible via BEC client
    USER_ACCESS = ["sim", "readback", "speed", "dummy_controller", "registered_proxies"]

    sim_cls = SimulatedPositioner

    # Define the signals as class attributes
    readback = Cpt(ReadOnlySignal, name="readback", value=0, kind=Kind.hinted)
    setpoint = Cpt(SetableSignal, value=0, kind=Kind.normal)
    motor_is_moving = Cpt(SetableSignal, value=0, kind=Kind.normal)

    # Config signals
    velocity = Cpt(SetableSignal, value=100, kind=Kind.config)
    acceleration = Cpt(SetableSignal, value=1, kind=Kind.config)

    # Ommitted signals
    high_limit_travel = Cpt(SetableSignal, value=0, kind=Kind.omitted)
    low_limit_travel = Cpt(SetableSignal, value=0, kind=Kind.omitted)
    unused = Cpt(SetableSignal, value=1, kind=Kind.omitted)

    SUB_READBACK = "readback"
    _default_sub = SUB_READBACK

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        name,
        *,
        delay: int = 1,
        update_frequency=2,
        precision=3,
        tolerance: float = 0.5,
        limits=None,
        parent=None,
        kind=None,
        device_manager=None,
        sim_init: dict = None,
        # TODO remove after refactoring config
        speed: float = 100,
        **kwargs,
    ):
        self.delay = delay
        self.device_manager = device_manager
        self.precision = precision
        self.tolerance = tolerance
        self.init_sim_params = sim_init
        self._registered_proxies = {}

        self.update_frequency = update_frequency
        self._stopped = False
        self.dummy_controller = DummyController()

        self.sim = self.sim_cls(parent=self, **kwargs)

        super().__init__(name=name, parent=parent, kind=kind, **kwargs)
        self.sim.sim_state[self.name] = self.sim.sim_state.pop(self.readback.name, None)
        self.readback.name = self.name
        if limits is not None:
            assert len(limits) == 2
            self.low_limit_travel.put(limits[0])
            self.high_limit_travel.put(limits[1])

    # @property
    # def connected(self):
    #     """Return the connected state of the simulated device."""
    #     return self.dummy_controller.connected

    @property
    def limits(self):
        """Return the limits of the simulated device."""
        return (self.low_limit_travel.get(), self.high_limit_travel.get())

    @property
    def low_limit(self):
        """Return the low limit of the simulated device."""
        return self.limits[0]

    @property
    def high_limit(self):
        """Return the high limit of the simulated device."""
        return self.limits[1]

    def registered_proxies(self) -> None:
        """Dictionary of registered signal_names and proxies."""
        return self._registered_proxies

    # pylint: disable=arguments-differ
    def check_value(self, value: any):
        """
        Check that requested position is within existing limits.

        This function has to be implemented on the top level of the positioner.
        """
        low_limit, high_limit = self.limits

        if low_limit < high_limit and not low_limit <= value <= high_limit:
            raise LimitError(f"position={value} not within limits {self.limits}")

    def _set_sim_state(self, signal_name: str, value: any) -> None:
        """Update the simulated state of the device."""
        self.sim.sim_state[signal_name]["value"] = value
        self.sim.sim_state[signal_name]["timestamp"] = ttime.time()

    def _get_sim_state(self, signal_name: str) -> any:
        """Return the simulated state of the device."""
        return self.sim.sim_state[signal_name]["value"]

    def move(self, value: float, **kwargs) -> DeviceStatus:
        """Change the setpoint of the simulated device, and simultaneously initiated a motion."""
        self._stopped = False
        self.check_value(value)
        old_setpoint = self._get_sim_state(self.setpoint.name)
        self._set_sim_state(self.motor_is_moving.name, 1)
        self._set_sim_state(self.setpoint.name, value)

        def update_state(val):
            """Update the state of the simulated device."""
            if self._stopped:
                raise DeviceStop
            old_readback = self._get_sim_state(self.readback.name)
            self._set_sim_state(self.readback.name, val)

            # Run subscription on "readback"
            self._run_subs(
                sub_type=self.SUB_READBACK,
                old_value=old_readback,
                value=self.sim.sim_state[self.readback.name]["value"],
                timestamp=self.sim.sim_state[self.readback.name]["timestamp"],
            )

        st = DeviceStatus(device=self)
        if self.delay:

            def move_and_finish():
                """Move the simulated device and finish the motion."""
                success = True
                try:
                    move_val = self._get_sim_state(
                        self.setpoint.name
                    ) + self.tolerance * np.random.uniform(-1, 1)

                    updates = np.ceil(
                        np.abs(old_setpoint - move_val)
                        / self.velocity.get()
                        * self.update_frequency
                    )

                    for ii in np.linspace(old_setpoint, move_val, int(updates)):
                        ttime.sleep(1 / self.update_frequency)
                        update_state(ii)

                    update_state(move_val)
                    self._set_sim_state(self.motor_is_moving, 0)
                except DeviceStop:
                    success = False
                finally:
                    self._stopped = False
                self._done_moving(success=success)
                self._set_sim_state(self.motor_is_moving.name, 0)
                st.set_finished()

            threading.Thread(target=move_and_finish, daemon=True).start()

        else:
            update_state(value)
            self._done_moving()
            self._set_sim_state(self.motor_is_moving.name, 0)
            st.set_finished()
        return st

    def stop(self, *, success=False):
        """Stop the motion of the simulated device."""
        super().stop(success=success)
        self._stopped = True

    @property
    def position(self) -> float:
        """Return the current position of the simulated device."""
        return self.readback.get()

    @property
    def egu(self):
        """Return the engineering units of the simulated device."""
        return "mm"


class SimFlyer(Device, PositionerBase, FlyerInterface):
    """A simulated device mimicing any 2D Flyer device (position, temperature, rotation).

    The corresponding simulation class is sim_cls=SimulatedPositioner, more details on defaults within the simulation class.

    >>> flyer = SimFlyer(name="flyer")

    Parameters
    ----------
    name (string)           : Name of the device. This is the only required argmuent, passed on to all signals of the device.
    precision (integer)     : Precision of the readback in digits, written to .describe(). Default is 3 digits.
    parent                  : Parent device, optional, is used internally if this signal/device is part of a larger device.
    kind                    : A member the Kind IntEnum (or equivalent integer), optional. Default is Kind.normal. See Kind for options.
    device_manager          : DeviceManager from BEC, optional . Within startup of simulation, device_manager is passed on automatically.
    """

    USER_ACCESS = ["sim", "registered_proxies"]

    sim_cls = SimulatedPositioner

    readback = Cpt(
        ReadOnlySignal, name="readback", value=0, kind=Kind.hinted, compute_readback=False
    )

    def __init__(
        self,
        name: str,
        *,
        precision: int = 3,
        parent=None,
        kind=None,
        device_manager=None,
        # TODO remove after refactoring config
        speed: float = 100,
        delay: int = 1,
        update_frequency: int = 100,
        **kwargs,
    ):

        self.sim = self.sim_cls(parent=self, device_manager=device_manager, **kwargs)
        self.precision = precision
        self.device_manager = device_manager
        self._registered_proxies = {}

        super().__init__(name=name, parent=parent, kind=kind, **kwargs)
        self.sim.sim_state[self.name] = self.sim.sim_state.pop(self.readback.name, None)
        self.readback.name = self.name

    @property
    def registered_proxies(self) -> None:
        """Dictionary of registered signal_names and proxies."""
        return self._registered_proxies

    @property
    def hints(self):
        """Return the hints of the simulated device."""
        return {"fields": ["flyer_samx", "flyer_samy"]}

    @property
    def egu(self) -> str:
        """Return the engineering units of the simulated device."""
        return "mm"

    def complete(self) -> StatusBase:
        """Complete the motion of the simulated device."""
        status = DeviceStatus(self)
        status.set_finished()
        return status

    def kickoff(self, metadata, num_pos, positions, exp_time: float = 0):
        """Kickoff the flyer to execute code during the scan."""
        positions = np.asarray(positions)

        def produce_data(device, metadata):
            """Simulate the data being produced by the flyer."""
            buffer_time = 0.2
            elapsed_time = 0
            bundle = messages.BundleMessage()
            for ii in range(num_pos):
                bundle.append(
                    messages.DeviceMessage(
                        signals={
                            "flyer_samx": {"value": positions[ii, 0], "timestamp": 0},
                            "flyer_samy": {"value": positions[ii, 1], "timestamp": 0},
                        },
                        metadata={"point_id": ii, **metadata},
                    )
                )
                ttime.sleep(exp_time)
                elapsed_time += exp_time
                if elapsed_time > buffer_time:
                    elapsed_time = 0
                    logger.info(f"Sending data point {ii} for {device.name}.")
                    device.device_manager.connector.set_and_publish(
                        MessageEndpoints.device_read(device.name), bundle
                    )
                    bundle = messages.BundleMessage()
                    device.device_manager.connector.set(
                        MessageEndpoints.device_status(device.name),
                        messages.DeviceStatusMessage(
                            device=device.name, status=1, metadata={"point_id": ii, **metadata}
                        ),
                    )
            device.device_manager.connector.set_and_publish(
                MessageEndpoints.device_read(device.name), bundle
            )
            device.device_manager.connector.set(
                MessageEndpoints.device_status(device.name),
                messages.DeviceStatusMessage(
                    device=device.name, status=0, metadata={"point_id": num_pos, **metadata}
                ),
            )
            print("done")

        flyer = threading.Thread(target=produce_data, args=(self, metadata))
        flyer.start()


class SynDeviceSubOPAAS(Device):
    zsub = Cpt(SimPositioner, name="zsub")


class SynDeviceOPAAS(Device):
    x = Cpt(SimPositioner, name="x")
    y = Cpt(SimPositioner, name="y")
    z = Cpt(SynDeviceSubOPAAS, name="z")


class SynDynamicComponents(Device):
    messages = Dcpt({f"message{i}": (SynSignal, None, {"name": f"msg{i}"}) for i in range(1, 6)})


class SimPositionerWithCommFailure(SimPositioner):
    fails = Cpt(SetableSignal, value=0)

    def move(self, value: float, **kwargs) -> DeviceStatus:
        if self.fails.get() == 1:
            raise RuntimeError("Communication failure")
        if self.fails.get() == 2:
            while not self._stopped:
                ttime.sleep(1)
            status = DeviceStatus(self)
            status.set_exception(RuntimeError("Communication failure"))
        return super().move(value, **kwargs)


if __name__ == "__main__":
    waveform = SimWaveform(name="waveform")
    waveform.sim.sim_select_model(waveform.sim.sim_get_models()[7])
    waveform.sim.sim_params = {
        "amplitude": 1500,
        "noise_multiplier": 168,
        "sigma": 50,
        "center": 350,
    }
    waveform.waveform.get()



from bec_lib import ServiceConfig
from bec_lib.messages import BECStatus
from bec_lib.tests.utils import ConnectorMock
from bec_server.scan_server.scan_server import ScanServer
from bec_server.scan_server.scan_worker import InstructionQueueStatus

# pylint: disable=missing-function-docstring
# pylint: disable=protected-access


class WorkerMock:
    def __init__(self) -> None:
        self.scan_id = None
        self.scan_motors = []
        self.current_scan_id = None
        self.current_scan_info = None
        self.status = InstructionQueueStatus.IDLE
        self.current_instruction_queue_item = None


class ScanServerMock(ScanServer):
    def __init__(self, device_manager) -> None:
        self.device_manager = device_manager
        super().__init__(
            ServiceConfig(redis={"host": "dummy", "port": 6379}), connector_cls=ConnectorMock
        )
        self.scan_worker = WorkerMock()

    def _start_metrics_emitter(self):
        pass

    def _start_update_service_info(self):
        pass

    def _start_device_manager(self):
        pass

    def shutdown(self):
        pass

    def wait_for_service(self, name, status=BECStatus.RUNNING):
        pass

    @property
    def scan_number(self) -> int:
        """get the current scan number"""
        return 2

    @scan_number.setter
    def scan_number(self, val: int):
        pass

    @property
    def dataset_number(self) -> int:
        """get the current dataset number"""
        return 3

    @dataset_number.setter
    def dataset_number(self, val: int):
        pass



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import datetime
import pathlib

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information
import tomli

project = "BEC"
copyright = f"{datetime.datetime.today().year}, Paul Scherrer Institute, Switzerland"
author = "Klaus Wakonig"

current_path = pathlib.Path(__file__).parent.parent.parent.resolve()
version_path = f"{current_path}/bec_server/pyproject.toml"


def get_version():
    """load the version from the version file"""
    with open(version_path, "r", encoding="utf-8") as file:
        res = tomli.loads(file.read())
    return res["project"]["version"]


release = get_version()


extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    # "sphinx.ext.coverage",
    "sphinx.ext.viewcode",
    "sphinx.ext.napoleon",
    "sphinx_toolbox.collapse",
    "sphinx_copybutton",
    "myst_parser",
    "sphinx_design",
]

myst_enable_extensions = [
    "amsmath",
    "attrs_inline",
    "colon_fence",
    "deflist",
    "dollarmath",
    "fieldlist",
    "html_admonition",
    "html_image",
    "replacements",
    "smartquotes",
    "strikethrough",
    "substitution",
    "tasklist",
]

autosummary_generate = True  # Turn on sphinx.ext.autosummary
add_module_names = False  # Remove namespaces from class/method signatures
autodoc_inherit_docstrings = True  # If no docstring, inherit from base class
set_type_checking_flag = True  # Enable 'expensive' imports for sphinx_autodoc_typehints
autoclass_content = "both"  # Include both class docstring and __init__

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

language = "Python"

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "pydata_sphinx_theme"
html_static_path = ["_static"]
html_css_files = ["css/custom.css"]
html_logo = "_static/bec.png"



import os
from pathlib import Path
from unittest import mock

import pytest

from bec_lib.bec_errors import ServiceConfigError
from bec_lib.logger import BECLogger
from bec_lib.tests.utils import ConnectorMock


@pytest.fixture
def logger():
    BECLogger._reset_singleton()
    logger = BECLogger()
    yield logger


def test_configure(logger, tmp_path):
    with mock.patch.object(logger, "_update_base_path") as mock_update_base:
        with mock.patch.object(logger, "writer_mixin") as mock_writer_mixin:
            with mock.patch.object(logger, "_update_sinks") as mock_update_sinks:
                logger._base_path = tmp_path
                logger.configure(
                    bootstrap_server=["localhost:9092"],
                    connector_cls=ConnectorMock,
                    service_name="test",
                    service_config={"log_writer": {"base_path": f"{tmp_path}"}},
                )
                assert mock_update_base.called is False
                assert mock_writer_mixin.called is False
                assert mock_update_sinks.mock_calls == mock.call
                assert logger.bootstrap_server == ["localhost:9092"]
                assert logger.service_name == "test"
                assert logger._configured is True


def test_update_base_path_correct_config(logger):
    config = {"log_writer": {"base_path": "./"}}
    assert logger._base_path is None
    logger._update_base_path(config)
    assert logger._base_path == os.path.join(str(Path("./").resolve()), "logs")


def test_update_base_path_wrong_config(logger):
    config = {"file_writer": {"base_path": "./"}}
    assert logger._base_path is None
    with pytest.raises(ServiceConfigError):
        logger._update_base_path(config)
